Package: resourcecode
Title: Access to the 'RESOURCECODE' Hindcast Database
Version: 0.5.4
Authors@R: 
    person("Nicolas", "Raillard", , "nicolas.raillard@ifremer.fr", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-3385-5104"))
Description: Utility functions to download data from the 'RESOURCECODE'
    hindcast database of sea-states, time series of sea-state parameters
    and time series of 1D and 2D wave spectra.  See
    <https://resourcecode.ifremer.fr> for more details about the available
    data.  Also provides facilities to plot and analyse downloaded data,
    such as computing the sea-state parameters from both the 1D and 2D
    surface elevation variance spectral density.
License: GPL (>= 3)
URL: https://github.com/Resourcecode-project/r-resourcecode,
        https://resourcecode-project.r-universe.dev/resourcecode
BugReports: https://github.com/Resourcecode-project/r-resourcecode/issues
Depends: R (>= 4.1)
Imports: abind, geosphere, ggplot2, grid, gridtext, httr2, lubridate,
        ncdf4, patchwork, pracma, Rcpp, resourcecodedata, rlang, sf,
        stats, tibble, tidyr
Suggests: knitr, mockery, rmarkdown, testthat, vcr
LinkingTo: Rcpp, RcppArmadillo
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Config/testthat/parallel: true
Encoding: UTF-8
Language: en-US
LazyData: true
NeedsCompilation: yes
Packaged: 2026-05-06 20:06:56 UTC; nraillar
Author: Nicolas Raillard [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-3385-5104>)
Maintainer: Nicolas Raillard <nicolas.raillard@ifremer.fr>
Repository: CRAN
Date/Publication: 2026-05-07 09:40:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-05-07 11:42:45 UTC; unix
Archs: resourcecode.so.dSYM

Package: resourcecode
Title: Access to the 'RESOURCECODE' Hindcast Database
Version: 0.5.5
Authors@R: 
    person("Nicolas", "Raillard", , "nicolas.raillard@ifremer.fr", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-3385-5104"))
Description: Utility functions to download data from the 'RESOURCECODE'
    hindcast database of sea-states, time series of sea-state parameters
    and time series of 1D and 2D wave spectra.  See
    <https://resourcecode.ifremer.fr> for more details about the available
    data.  Also provides facilities to plot and analyse downloaded data,
    such as computing the sea-state parameters from both the 1D and 2D
    surface elevation variance spectral density.
License: GPL (>= 3)
URL: https://github.com/Resourcecode-project/r-resourcecode,
        https://resourcecode-project.r-universe.dev/resourcecode
BugReports: https://github.com/Resourcecode-project/r-resourcecode/issues
Depends: R (>= 4.1)
Imports: abind, geosphere, ggplot2, grid, gridtext, httr2, lubridate,
        ncdf4, patchwork, pracma, Rcpp, resourcecodedata, rlang, sf,
        stats, tibble, tidyr
Suggests: knitr, mockery, rmarkdown, testthat, vcr
LinkingTo: Rcpp, RcppArmadillo
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Config/testthat/parallel: true
Encoding: UTF-8
Language: en-US
LazyData: true
NeedsCompilation: yes
Packaged: 2026-08-03 13:09:03 UTC; nraillar
Author: Nicolas Raillard [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-3385-5104>)
Maintainer: Nicolas Raillard <nicolas.raillard@ifremer.fr>
Repository: CRAN
Date/Publication: 2026-08-03 23:20:02 UTC

# Generated by roxygen2: do not edit by hand

export("%nin%")
export(closest_point_field)
export(closest_point_spec)
export(compute_orbital_speeds)
export(compute_sea_state_1d_spectrum)
export(compute_sea_state_2d_spectrum)
export(convert_spectrum_2d1d)
export(cut_directions)
export(cut_seasons)
export(dispersion)
export(fastTrapz)
export(fractional_day_of_year)
export(get_1d_spectrum)
export(get_2d_spectrum)
export(get_parameters)
export(jonswap)
export(mean_direction)
export(metconv2zmcomp)
export(plot_1d_specta)
export(plot_2d_specta)
export(rscd_mapplot)
export(weather_windows)
export(zmcomp2metconv)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(ggplot2,aes)
importFrom(ggplot2,coord_polar)
importFrom(ggplot2,coord_sf)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,expansion)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_path)
importFrom(ggplot2,geom_polygon)
importFrom(ggplot2,geom_rect)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,labs)
importFrom(ggplot2,margin)
importFrom(ggplot2,scale_color_distiller)
importFrom(ggplot2,scale_fill_distiller)
importFrom(ggplot2,scale_x_continuous)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_linedraw)
importFrom(ggplot2,theme_void)
importFrom(rlang,.data)
useDynLib(resourcecode, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# Resource**code** <a href="https://github.com/Resourcecode-project/r-resourcecode"><img src="man/figures/logo.png" align="right" height="139" alt="resourcecode website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/Resourcecode-project/r-resourcecode/actions/workflows/R-CMD-check.yaml/badge.svg?branch=main)](https://github.com/Resourcecode-project/r-resourcecode/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/Resourcecode-project/r-resourcecode/branch/main/graph/badge.svg)](https://app.codecov.io/gh/Resourcecode-project/r-resourcecode)
[![resourcecode status
badge](https://resourcecode-project.r-universe.dev/resourcecode/badges/version)](https://resourcecode-project.r-universe.dev/resourcecode)
[![CRAN
status](https://www.r-pkg.org/badges/version/resourcecode)](https://CRAN.R-project.org/package=resourcecode)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/resourcecode)](https://CRAN.R-project.org/package=resourcecode)
<!-- badges: end -->

The goal of `{resourcecode}` is to provide an easy access to the
Resourcecode hindcast database of sea-states. More information on the
database can be found [here](https://resourcecode.ifremer.fr). This
website only contains information about the way to retrieve data from
this data base and the functionalities offred by this package.

## Installation

The `{resourcecode}` package is on CRAN so you can simply run :

``` r
install.packages("resourcecode")
```

You can install the development version of `{resourcecode}` thanks to
[r-universe](https://resourcecode-project.r-universe.dev/resourcecode):

``` r
install.packages("resourcecode",
  repos = c(
    "https://resourcecode-project.r-universe.dev",
    "https://cloud.r-project.org"
  )
)
```

or using the classical:

``` r
devtools::install_github("Resourcecode-project/r-resourcecode")
```

This package depends on data in a data package `{resourcecodedata}` that
is available on CRAN on will be installed when installing the package
for the first time.

## Examples

Plot the bathymetry used in the project

``` r
library(resourcecode)
library(resourcecodedata)
resourcecode::rscd_mapplot(rscd_field$depth, name = "Depth (m)", transform = "sqrt")
```

<img src="man/figures/README-bathymetry-1.png" alt="" width="100%" style="display: block; margin: auto;" />

See the variables available in the database:

``` r
head(rscd_variables)
#>        name         longname         unit
#> 1 longitude        longitude  degree_east
#> 2  latitude         latitude degree_north
#> 3       tri              tri             
#> 4    MAPSTA       status map            1
#> 5       dpt            depth            m
#> 6      ucur eastward current        m s-1
```

Download a time series of significant wave height next to the coast of
Finistère:

``` r
data <- get_parameters(node = "134865", parameters = "hs")
str(data)
#> tibble [8,760 × 2] (S3: tbl_df/tbl/data.frame)
#>  $ time: POSIXct[1:8760], format: "1994-01-01 00:00:00" "1994-01-01 01:00:00" ...
#>  $ hs  : num [1:8760] 4.82 4.98 5.19 5.37 5.48 ...
#>  - attr(*, "node")= num 134864
if (!is.null(data)) plot(data, type = "l")
```

<img src="man/figures/README-data_fetcher-1.png" alt="" width="100%" style="display: block; margin: auto;" />

#' Plot a wave density 2D spectrum
#'
#' @param spec the spectral data, as an output from `get_2Dspectrum`
#' @param time the time to plot. Either an integer or the date.
#' @param normalize Should the spectrum be normalized to have maximum 1 before plotting
#' @param trim removes the values of the spectral density lower than this value
#' @param cut_off cut-off frequency above which the spectrum is not plotted
#' @param ... currently unused
#'
#' @return a ggplot object
#' @export
#'
#' @examples
#' plot_2d_specta(resourcecodedata::rscd_2d_spectra, 1)
#' @importFrom ggplot2 ggplot geom_rect scale_x_continuous scale_y_continuous
#' @importFrom ggplot2 theme_linedraw coord_polar scale_color_distiller scale_fill_distiller
#' @importFrom ggplot2 labs expansion
plot_2d_specta <- function(
  spec,
  time = 1L,
  normalize = TRUE,
  trim = 0.01,
  cut_off = 0.4,
  ...
) {
  if (is.character(time)) {
    time <- as.POSIXct(time, tz = "UTC")
  }

  if (inherits(time, "POSIXct")) {
    time <- which(time == spec$forcings$time)
  }

  df_freq <- tidyr::expand_grid(frequency1 = 1:36, dir = spec$dir)

  df_freq$frequency2 <- spec$frequency2[df_freq$frequency1]
  df_freq$frequency1 <- spec$frequency1[df_freq$frequency1]

  df <- tibble::tibble(df_freq, ef = c(spec$efth[,, time]))

  legend_text <- "Power spectrum\n(m^2.s)"

  if (normalize) {
    ef_max <- max(df$ef)
    df$ef <- df$ef / ef_max

    legend_text <- "Normalized\nPower spectrum"
  }

  if (!is.null(trim)) {
    # Trim
    df$ef[df$ef <= trim] <- NA
  }

  df$dir <- (df$dir + 180) %% 360

  ggplot(
    df,
    aes(
      xmin = .data$frequency1,
      xmax = .data$frequency2,
      ymin = dir - 5,
      ymax = dir + 5,
      fill = .data$ef,
      col = .data$ef
    )
  ) +
    geom_rect() +
    scale_x_continuous(
      name = "Frequency (Hz)",
      expand = expansion(),
      limits = c(0, cut_off)
    ) +
    scale_y_continuous(
      name = "Direction from (\u00b0)",
      expand = expansion(),
      breaks = c(0, 90, 180, 270),
      labels = c("N", "E", "S", "W"),
      minor_breaks = seq(from = 0, to = 360, by = 30)
    ) +
    coord_polar(theta = "y", start = -5 * pi / 180) +
    scale_color_distiller(
      palette = "PuBu",
      direction = 1,
      name = legend_text,
      na.value = "transparent"
    ) +
    scale_fill_distiller(
      palette = "PuBu",
      direction = 1,
      name = legend_text,
      na.value = "transparent"
    ) +
    labs(
      title = paste(
        "Directional Wave energy spectrum at location",
        spec$station
      ),
      subtitle = format(spec$forcings$time[time], format = "%Y-%m-%d %H:%M"),
      caption = "Source: Resourcecode hindcast database\nresourcecode.ifremer.fr"
    ) +
    theme_linedraw() +
    theme(legend.position = "bottom")
}

#' Plot a wave density 1D spectrum at a given time
#'
#' @param spec the spectral data, as an output from `get_2Dspectrum`
#' @param time the time to plot. Either an integer or the date.
#' @param print_sea_state should the sea_states parameters being plot ? Default to TRUE.
#' @param ... currently unused
#'
#' @return a ggplot object
#' @export
#'
#' @examples
#' plot_1d_specta(resourcecodedata::rscd_1d_spectra, 1)
#' @importFrom ggplot2 ggplot geom_line scale_x_continuous scale_y_continuous
#' @importFrom ggplot2 theme_linedraw labs expansion
plot_1d_specta <- function(
  spec,
  time = 1L,
  print_sea_state = TRUE,
  ...
) {
  if (is.character(time)) {
    time <- as.POSIXct(time, tz = "UTC")
  }

  if (inherits(time, "POSIXct")) {
    time <- which(time == spec$forcings$time)
  }

  if ("freq" %nin% names(spec)) {
    warning(
      "Frequency vector not provided: using the default Resourcecode frequency vector."
    )
    freq_plot <- resourcecodedata::rscd_freq
  } else {
    freq_plot <- spec$freq
  }

  df_plot <- tibble::tibble(
    freq = freq_plot,
    spectral_density = spec$ef[, time]
  )

  out_plot <- ggplot(
    df_plot,
    aes(x = .data$freq, y = .data$spectral_density)
  ) +
    geom_line(col = "#045A8D") + # Colour is the darkest from 'PuBu' for consistency with 2d plot
    theme_linedraw() +
    scale_x_continuous(
      name = "Frequency (Hz)",
      breaks = seq(from = 0, to = 1, by = 0.2),
      expand = FALSE
    ) +
    scale_y_continuous(
      name = "Wave spectral density (m^2.s)",
      expand = expansion(c(0, .01), c(0, .05))
    ) +
    labs(
      title = paste("Wave energy spectrum at location", spec$station),
      subtitle = format(spec$forcings$time[time], format = "%Y-%m-%d %H:%M"),
      caption = "Source: Resourcecode hindcast database\nresourcecode.ifremer.fr"
    )
  if (print_sea_state) {
    text <- paste0(
      "**Sea-state parameters:**\n
Hs=",
      round(spec$forcings$hs[time], 2),
      "m<br>
Tp=",
      round(1 / spec$forcings$fp[time], 2),
      "s<br>
Dir=",
      round(spec$forcings$dir[time], 2),
      "\u00b0<br>
Wspd=",
      round(spec$forcings$wnd[time], 2),
      "m/s<br>
Wdir=",
      round(spec$forcings$wnddir[time], 2),
      "\u00b0<br>"
    )

    out_plot <- out_plot +
      patchwork::inset_element(
        gridtext::textbox_grob(
          text,
          box_gp = grid::gpar(col = "black"),
          r = grid::unit(5, "pt"),
          padding = grid::unit(c(10, 0, 0, 5), "pt"),
          margin = grid::unit(c(0, 0, 0, 0), "pt")
        ),
        left = 0.7,
        bottom = 0.77,
        right = .95,
        top = .95
      )
  }
  out_plot
}

test_that("we can plot the 2D spectrum", {
  spec <- resourcecodedata::rscd_2d_spectra
  plot_from_index <- plot_2d_specta(spec, 1)
  plot_not_normalized <- plot_2d_specta(spec, 1, normalize = FALSE)
  plot_from_date <- plot_2d_specta(spec, "1994-01-27 13:00:00")
  expect_no_warning(print(plot_from_index))
  expect_no_warning(print(plot_not_normalized))
  expect_no_warning(print(plot_from_date))
  # vdiffr::expect_doppelganger(
  #   "Specifying index",
  #   plot_from_index,
  #   variant = Sys.info()[["sysname"]]
  # )
  # vdiffr::expect_doppelganger(
  #   "Un-normalized 2D spectra",
  #   plot_not_normalized,
  #   variant = Sys.info()[["sysname"]]
  # )
  # vdiffr::expect_doppelganger(
  #   "Plot by date",
  #   plot_from_date,
  #   variant = Sys.info()[["sysname"]]
  # )
})

test_that("Plotting maps works", {
  expect_no_error(rscd_mapplot(
    resourcecodedata::rscd_field$depth,
    name = "Depth (m)"
  ))
  expect_no_warning(rscd_mapplot(
    resourcecodedata::rscd_field$depth,
    name = "Depth (m)"
  ))
  # vdiffr::expect_doppelganger(
  #   "Mapping water depth",
  #   rscd_mapplot(resourcecodedata::rscd_field$depth, name = "Depth (m)"),
  #   variant = Sys.info()[["sysname"]]
  # )
})

# Test file: test-plot_1d_specta.R

test_that("plot_1d_specta returns a ggplot object", {
  result <- plot_1d_specta(resourcecodedata::rscd_1d_spectra, time = 1)

  expect_s3_class(result, "ggplot")
})

test_that("plot_1d_specta works with integer time input", {
  result <- plot_1d_specta(resourcecodedata::rscd_1d_spectra, time = 5L)

  expect_s3_class(result, "ggplot")
})

test_that("plot_1d_specta works with character time input", {
  # Get a valid time from the data
  valid_time <- format(
    resourcecodedata::rscd_1d_spectra$forcings$time[3],
    format = "%Y-%m-%d %H:%M:%S"
  )

  result <- plot_1d_specta(resourcecodedata::rscd_1d_spectra, time = valid_time)

  expect_s3_class(result, "ggplot")
})

test_that("plot_1d_specta works with POSIXct time input", {
  # Use an actual time from the data
  time_posix <- resourcecodedata::rscd_1d_spectra$forcings$time[2]

  result <- plot_1d_specta(resourcecodedata::rscd_1d_spectra, time = time_posix)

  expect_s3_class(result, "ggplot")
})

test_that("plot_1d_specta warns when freq vector not provided", {
  # Create a copy without the freq vector
  spec_no_freq <- resourcecodedata::rscd_1d_spectra
  spec_no_freq$freq <- NULL

  expect_warning(
    plot_1d_specta(spec_no_freq, time = 1),
    "Frequency vector not provided: using the default Resourcecode frequency vector."
  )
})

test_that("plot_1d_specta works with print_sea_state = FALSE", {
  result <- plot_1d_specta(
    resourcecodedata::rscd_1d_spectra,
    time = 1,
    print_sea_state = FALSE
  )

  expect_s3_class(result, "ggplot")
})

test_that("plot_1d_specta works with print_sea_state = TRUE", {
  result <- plot_1d_specta(
    resourcecodedata::rscd_1d_spectra,
    time = 1,
    print_sea_state = TRUE
  )

  expect_s3_class(result, "ggplot")
  # When print_sea_state = TRUE, the plot should have additional components from patchwork
  expect_true(length(result) > 0)
})

test_that("plot_1d_specta has correct plot labels", {
  result <- plot_1d_specta(
    resourcecodedata::rscd_1d_spectra,
    time = 1,
    print_sea_state = FALSE
  )

  # Check that title contains station name
  expect_match(
    result@labels$title,
    paste(
      "Wave energy spectrum at location",
      resourcecodedata::rscd_1d_spectra$station
    )
  )
  expect_equal(result@scales$get_scales("x")$name, "Frequency (Hz)")
  expect_equal(
    result@scales$get_scales("y")$name,
    "Wave spectral density (m^2.s)"
  )
  expect_match(result@labels$caption, "Source: Resourcecode hindcast database")
})

test_that("plot_1d_spectra handles edge case with first time point", {
  result <- plot_1d_specta(resourcecodedata::rscd_1d_spectra, time = 1)

  expect_s3_class(result, "ggplot")
})

test_that("plot_1d_spectra handles edge case with last time point", {
  # Get the last time index
  last_time <- ncol(resourcecodedata::rscd_1d_spectra$ef)

  result <- plot_1d_specta(resourcecodedata::rscd_1d_spectra, time = last_time)

  expect_s3_class(result, "ggplot")
})

test_that("plot_1d_specta subtitle contains correct time format", {
  result <- plot_1d_specta(
    resourcecodedata::rscd_1d_spectra,
    time = 1,
    print_sea_state = FALSE
  )

  # Check subtitle matches expected datetime format (YYYY-MM-DD HH:MM)
  expect_match(result$labels$subtitle, "^\\d{4}-\\d{2}-\\d{2} \\d{2}:\\d{2}$")
})

test_that("plot_1d_specta uses correct frequency data", {
  result <- plot_1d_specta(
    resourcecodedata::rscd_1d_spectra,
    time = 1,
    print_sea_state = FALSE
  )

  # Extract the data from the plot
  plot_data <- ggplot2::layer_data(result, 1)
  dim(plot_data$x) <- 36

  # Check that the frequency data matches the input
  expect_equal(plot_data$x, resourcecodedata::rscd_1d_spectra$freq)
})

test_that("plot_1d_specta uses correct spectral density data", {
  time_idx <- 3
  result <- plot_1d_specta(
    resourcecodedata::rscd_1d_spectra,
    time = time_idx,
    print_sea_state = FALSE
  )

  # Extract the data from the plot
  plot_data <- ggplot2::layer_data(result, 1)

  # Check that the spectral density data matches the input
  expect_equal(plot_data$y, resourcecodedata::rscd_1d_spectra$ef[, time_idx])
})

test_that("plot_1d_specta time conversion works correctly", {
  # Test with character time that exists in the data
  char_time <- format(
    resourcecodedata::rscd_1d_spectra$forcings$time[4],
    format = "%Y-%m-%d %H:%M:%S"
  )
  result_char <- plot_1d_specta(
    resourcecodedata::rscd_1d_spectra,
    time = char_time,
    print_sea_state = FALSE
  )

  # Test with integer time
  result_int <- plot_1d_specta(
    resourcecodedata::rscd_1d_spectra,
    time = 4L,
    print_sea_state = FALSE
  )

  # Test with POSIXct time
  result_posix <- plot_1d_specta(
    resourcecodedata::rscd_1d_spectra,
    time = resourcecodedata::rscd_1d_spectra$forcings$time[4],
    print_sea_state = FALSE
  )

  # All three should produce the same subtitle (same time point)
  expect_equal(result_char$labels$subtitle, result_int$labels$subtitle)
  expect_equal(result_int$labels$subtitle, result_posix$labels$subtitle)
})

test_that("plot_1d_specta has correct color for line", {
  result <- plot_1d_specta(
    resourcecodedata::rscd_1d_spectra,
    time = 1,
    print_sea_state = FALSE
  )

  # Check that the line color is the expected PuBu color
  expect_equal(result$layers[[1]]$aes_params$colour, "#045A8D")
})

test_that("plot_1d_specta has correct theme", {
  result <- plot_1d_specta(
    resourcecodedata::rscd_1d_spectra,
    time = 1,
    print_sea_state = FALSE
  )

  # Check that theme_linedraw is applied
  expect_s3_class(result$theme, "theme")
})

#' Compute Weather Windows
#'
#' Computes and returns start date of each weather window, implemented in C++ for speed.
#'
#' @param valid_periods A data frame with a 'time' column (POSIXct).
#' @param window_length Minimum window duration (hours).
#' @param allow_overlap Logical; If TRUE, the algorithm searches for window,
#' if a window is found, search of next window will start from the end of the previous window.
#' If FALSE, it uses continuous window search: The algorithm searches for window starting from
#' every time step that meets the criteria.
#' @param time_step Expected time step between consecutive timestamps (seconds).
#'
#' @return POSIXct vector of detected window start times.
#' @export
weather_windows <- function(
  valid_periods,
  window_length,
  allow_overlap = TRUE,
  time_step = 3600
) {
  if ("time" %nin% names(valid_periods)) {
    stop("'valid_periods' must be a data frame with a 'time' column (POSIXct)")
  }
  times_num <- as.numeric(valid_periods$time)
  detected <- ww_calc_cpp(times_num, window_length, allow_overlap, time_step)
  lubridate::as_datetime(detected)
}

# Helper: create hourly sequence with optional gaps
make_times <- function(start, end, by = "hour", gaps = NULL) {
  t <- seq.POSIXt(as.POSIXct(start), as.POSIXct(end), by = by)
  if (!is.null(gaps)) {
    t <- t[-gaps]
  }
  data.frame(time = t)
}

test_that("'weather_windows()' detects windows correctly with overlap", {
  df <- make_times("2020-01-01 00:00:00", "2020-01-02 23:00:00")

  # 24-hour windows, allow overlap
  res <- weather_windows(
    df,
    window_length = 24,
    allow_overlap = TRUE,
    time_step = 3600
  )

  expect_true(length(res) > 0)
  expect_true(all(res >= df$time[1]))
  expect_true(all(diff(res) >= 1)) # minimal check: diff positive
})

test_that("'weather_windows()' respects allow_overlap = FALSE", {
  df <- make_times("2020-01-01 00:00:00", "2020-01-03 23:00:00")

  res_overlap <- weather_windows(df, window_length = 24, allow_overlap = TRUE)
  res_no_overlap <- weather_windows(
    df,
    window_length = 24,
    allow_overlap = FALSE
  )

  expect_true(length(res_overlap) <= length(res_no_overlap))
  expect_true(all(diff(res_overlap, units = "hours") >= 24)) # no overlapping windows
})

test_that("'weather_windows()' handles gaps correctly", {
  # Remove some timestamps to simulate gaps
  df <- make_times(
    "2020-01-01 00:00:00",
    "2020-01-02 23:00:00",
    gaps = c(10, 11, 12)
  )

  res <- weather_windows(df, window_length = 5, allow_overlap = TRUE)
  # Ensure all returned starts are within the original data
  expect_true(all(res %in% df$time))
})

test_that("'weather_windows()' returns empty for insufficient data", {
  df <- data.frame(time = as.POSIXct("2020-01-01 00:00:00"))
  expect_equal(length(weather_windows(df, window_length = 1)), 0)

  df2 <- make_times("2020-01-01 00:00:00", "2020-01-01 01:00:00")
  # window_length longer than data
  expect_equal(length(weather_windows(df2, window_length = 5)), 0)
})

test_that("'weather_windows()' handles non-hourly time_step", {
  df <- make_times("2020-01-01 00:00:00", "2020-01-01 06:00:00", by = "2 hours")

  res <- weather_windows(
    df,
    window_length = 4,
    allow_overlap = TRUE,
    time_step = 2 * 3600
  )
  expect_true(length(res) > 0)
})

#' Create a map of the provided variable on the RESOURCECODE field grid
#'
#' @param z the data ro plot: a vector of the same size as the grid (328,030 rows).
#' @param name name of the variable plored, to be included in the legend.
#' @param zlim limits of the scale. See \code{\link[ggplot2]{continuous_scale}} for details.
#' @param palette If a string, will use that named palette.
#' See \code{\link[ggplot2]{scale_colour_brewer}} for other options.
#' @param direction Sets the order of colours in the scale.
#' See \code{\link[ggplot2]{scale_colour_brewer}} for details.
#' @param transform  Transformation to apply to the color scale.
#' See \code{\link[ggplot2]{continuous_scale}} for details.
#'
#' @return a ggplot2 object
#' @export
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' rscd_mapplot(resourcecodedata::rscd_field$depth)
#'
rscd_mapplot <- function(
  z,
  name = "Depth (m)",
  zlim = NULL,
  palette = "YlOrRd",
  direction = 1,
  transform = "identity"
) {
  xyzgz <- tibble::tibble(
    x = resourcecodedata::rscd_field$longitude[
      resourcecodedata::rscd_triangles
    ],
    y = resourcecodedata::rscd_field$latitude[resourcecodedata::rscd_triangles],
    z = z[resourcecodedata::rscd_triangles],
    g = rep(
      seq_len(
        ncol(resourcecodedata::rscd_triangles)
      ),
      each = nrow(resourcecodedata::rscd_triangles)
    )
  )

  ggplot(
    data = xyzgz,
    aes(
      .data$x,
      .data$y,
      group = .data$g,
      fill = .data$z,
      col = .data$z
    )
  ) +
    geom_polygon() +
    scale_fill_distiller(
      name = name,
      palette = palette,
      na.value = "transparent",
      limits = zlim,
      direction = direction,
      transform = transform
    ) +
    scale_color_distiller(
      guide = "none",
      palette = palette,
      na.value = "transparent",
      limits = zlim,
      direction = direction,
      transform = transform
    ) +
    geom_path(
      data = resourcecodedata::rscd_coastline,
      aes(x = .data$longitude, y = .data$latitude),
      linewidth = 0.2,
      inherit.aes = FALSE
    ) +
    geom_path(
      data = resourcecodedata::rscd_islands,
      aes(
        x = .data$longitude,
        y = .data$latitude,
        group = .data$ID
      ),
      linewidth = 0.2,
      inherit.aes = FALSE
    ) +
    coord_sf(expand = FALSE, crs = sf::st_crs(4326)) +
    theme_void() +
    labs(
      caption = "Source: Resourcecode hindcast database\nresourcecode.ifremer.fr"
    ) +
    theme(
      legend.position = "inside",
      legend.position.inside = c(0.8, 0.2),
      legend.margin = margin(
        t = 0,
        r = 0,
        b = 0,
        l = 0,
        unit = "pt"
      ),
      plot.margin = margin(
        t = 0,
        r = 0,
        b = 0,
        l = 0,
        unit = "pt"
      ),
      axis.text.x = element_blank(),
      axis.ticks.x = element_blank(),
      axis.text.y = element_blank(),
      axis.ticks.y = element_blank(),
      panel.border = element_rect(
        color = "black",
        fill = NA,
        linewidth = 1
      )
    )
}

#' Get time series of sea-state parameters
#'
#' No verification is made within this function, and thus it should not
#' be called directly by the user
#'
#' @param parameter variable to fetch from the database
#' @param node single integer with the node to get
#' @param start starting date (posixct)
#' @param end end date (posixct)
#'
#' @return a tibble with 2 columns and as many rows as needed
#' @noRd
get_parameters_raw <- function(
  parameter = "hs",
  node = 42,
  start = as.POSIXct("1994-01-01Z00:00:00", tz = "UTC"),
  end = as.POSIXct("1994-12-31Z23:00:00", tz = "UTC")
) {
  if (parameter == "tp") {
    single_parameter <- "fp"
  } else {
    single_parameter <- parameter
  }

  start_str <- strftime(start, format = "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")
  end_str <- strftime(end, format = "%Y-%m-%dT%H:%M:%SZ", tz = "UTC")

  # Cassandra database start indexing at 1, so decrements node number
  node <- node - 1

  request_url <- paste0(
    rcd_cassandra_url,
    "api/timeseries",
    "?parameter=",
    single_parameter,
    "&node=",
    node,
    "&start=",
    start_str,
    "&end=",
    end_str
  )

  # Try retrieving and parsing JSON using httr2
  resp <- tryCatch(
    httr2::request(request_url) |>
      httr2::req_error(is_error = \(resp) FALSE) |> # Don't auto-error on HTTP errors
      httr2::req_retry(max_tries = 3) |> # Retry transient failures
      httr2::req_timeout(30) |> # 30 second timeout
      httr2::req_user_agent("Resourcecode R package") |>
      httr2::req_perform(),
    httr2_failure = function(cnd) {
      message(
        "Network error: Could not connect to the remote resource. ",
        "The server may be unavailable."
      )
      NULL
    },
    error = function(e) {
      message("Unexpected error retrieving data: ", conditionMessage(e))
      NULL
    }
  )

  # If request failed, exit
  if (is.null(resp)) {
    return(NULL)
  }

  # Check HTTP status
  if (httr2::resp_status(resp) != 200) {
    message(
      "HTTP error ",
      httr2::resp_status(resp),
      ": ",
      httr2::resp_status_desc(resp)
    )
    return(NULL)
  }

  # Parse JSON response
  res <- tryCatch(
    httr2::resp_body_json(resp, simplifyVector = TRUE),
    error = function(e) {
      message("Error parsing response: Invalid JSON format")
      NULL
    }
  )

  # If parsing failed, exit
  if (is.null(res)) {
    return(NULL)
  }

  # Check API-level error
  if (!is.null(res$errorcode) && res$errorcode != 0) {
    message(
      "The data source returned an error: ",
      res$errormessage,
      "\nReturning NULL."
    )
    return(NULL) # graceful fallback
  }

  # Convert list to data frame
  data <- as.data.frame(res$result$data)
  colnames(data) <- c("time", parameter)
  data <- tibble::as_tibble(data)

  if (parameter == "tp") {
    data[, 2] <- 1 / data[, 2]
  }

  data$time <- as.POSIXct(
    as.numeric(data$time) / 1000,
    origin = as.POSIXct("1970-01-01", tz = "UTC"),
    tz = "UTC"
  ) # Convert UNIX time (ms) to POSIXct format
  attr(data, "node") <- node
  data
}


#' Download time series of sea-state parameters from RESOURCECODE database
#'
#' If the remote resource is unavailable or returns an error, the function returns NULL
#' and emits an informative message.
#'
#' @param parameters character vector of sea-state parameters
#' @param node single integer with the node to get
#' @param start starting date (as integer, character or posixct)
#' @param end ending date (as integer, character or posixct)
#'
#' @return a tibble with N-rows and `length(parameters)` columns.
#' @export
#'
#' @examples
#' rscd_data <- get_parameters(parameters = c("hs", "tp"), node = 42)
#' if(!is.null(rscd_data)) plot(rscd_data$time, rscd_data$hs, type = "l")
get_parameters <- function(
  parameters = "hs",
  node = 42,
  start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
  end = as.POSIXct("1994-12-31 23:00:00", tz = "UTC")
) {
  parameters <- tolower(parameters)

  if (any(parameters %nin% c("tp", resourcecodedata::rscd_variables$name))) {
    errors <- parameters[
      parameters %nin% c("tp", resourcecodedata::rscd_variables$name)
    ]
    stop(
      "Requested parameters do not exists in the database: ",
      paste0(errors, collapse = ", "),
      "."
    )
  }

  node <- as.integer(node)

  if (length(node) != 1) {
    stop("The function can retreive only one location a time.")
  }
  if (node %nin% resourcecodedata::rscd_field$node) {
    stop("The requested location do no exist in the database.")
  }

  if (is.character(start)) {
    start <- as.POSIXct(start, tz = "UTC")
  }
  if (is.character(end)) {
    end <- as.POSIXct(end, tz = "UTC")
  }

  if (is.numeric(start)) {
    start <- as.POSIXct(
      start,
      tz = "UTC",
      origin = as.POSIXct("1970-01-01 00:00:00", tz = "UTC")
    )
  }
  if (is.numeric(end)) {
    end <- as.POSIXct(
      end,
      tz = "UTC",
      origin = as.POSIXct("1970-01-01 00:00:00", tz = "UTC")
    )
  }
  if (lubridate::is.Date(start)) {
    start = as.POSIXct(paste(start, "00:00:00"), tz = "UTC")
  }
  if (lubridate::is.Date(end)) {
    end = as.POSIXct(paste(end, "23:00:00"), tz = "UTC")
  }
  if (start < rscd_casandra_start_date) {
    stop(
      "'start' is outside the covered period: ",
      paste(
        format(
          c(rscd_casandra_start_date, rscd_casandra_end_date),
          format = "%Y-%m-%d %H:%M %Z"
        ),
        collapse = " \u2014 "
      )
    )
  }
  if (end > rscd_casandra_end_date) {
    stop(
      "'end' is outside the covered period: ",
      paste(
        format(
          c(rscd_casandra_start_date, rscd_casandra_end_date),
          format = "%Y-%m-%d %H:%M %Z"
        ),
        collapse = " \u2014 "
      )
    )
  }
  if (start >= end) {
    stop("'end' must be after 'start'")
  }

  out <- get_parameters_raw(
    parameters[1],
    node = node,
    start = start,
    end = end
  )

  # If first parameter retrieval failed, return NULL
  if (is.null(out)) {
    message("Failed to retrieve parameter: ", parameters[1])
    return(NULL)
  }

  for (i in seq_len(length(parameters) - 1)) {
    temp <- get_parameters_raw(
      parameters[i + 1],
      node = node,
      start = start,
      end = end
    )

    # If any subsequent parameter retrieval fails, return NULL
    if (is.null(temp)) {
      message("Failed to retrieve parameter: ", parameters[i + 1])
      return(NULL)
    }

    out <- cbind.data.frame(out, temp[, 2])
  }
  out
}

# Configure vcr for your tests
# vcr_configure(
#   dir = "tests/fixtures/vcr_cassettes",
#   filter_sensitive_data = list(
#     # If you have API keys or sensitive data in URLs, filter them here
#     # "<<<MY_SECRET>>>" = Sys.getenv("SECRET_KEY")
#   )
# )

#Tests for get_parameters function (which also tests get_parameters_raw internally)
test_that("get_parameters retrieves single parameter and tests basic functionality", {
  vcr::local_cassette("get_single_parameter")
  result <- get_parameters(
    parameters = "hs",
    node = 42,
    start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
    end = as.POSIXct("1994-01-02 00:00:00", tz = "UTC")
  )

  expect_s3_class(result, "data.frame")
  expect_named(result, c("time", "hs"))
  expect_s3_class(result$time, "POSIXct")
  expect_type(result$hs, "double")
  expect_true(nrow(result) > 0)
})

test_that("get_parameters retrieves multiple parameters including tp conversion", {
  vcr::local_cassette("get_multiple_parameters")
  result <- get_parameters(
    parameters = c("hs", "tp"),
    node = 42,
    start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
    end = as.POSIXct("1994-01-02 00:00:00", tz = "UTC")
  )

  expect_s3_class(result, "data.frame")
  expect_named(result, c("time", "hs", "tp"))
  expect_true(nrow(result) > 0)
  # Check that tp values are positive (tests fp to tp conversion in get_parameters_raw)
  expect_true(all(result$tp > 0, na.rm = TRUE))
})

test_that("get_parameters handles character date inputs", {
  vcr::local_cassette("character_dates")
  result <- get_parameters(
    parameters = "hs",
    node = 42,
    start = "1994-01-01 00:00:00",
    end = "1994-01-02 00:00:00"
  )

  expect_s3_class(result, "data.frame")
  expect_true(nrow(result) > 0)
})

test_that("get_parameters handles Date format as input", {
  vcr::local_cassette("Date_dates")
  result <- get_parameters(
    parameters = "hs",
    node = 42,
    start = as.Date("1994-01-01"),
    end = as.Date("1994-01-01")
  )

  expect_s3_class(result, "data.frame")
  expect_true(nrow(result) > 0)
})

test_that("get_parameters handles numeric date inputs", {
  vcr::local_cassette("numeric_dates")
  start_num <- as.numeric(as.POSIXct("1994-01-01 00:00:00", tz = "UTC"))
  end_num <- as.numeric(as.POSIXct("1994-01-02 00:00:00", tz = "UTC"))

  result <- get_parameters(
    parameters = "hs",
    node = 42,
    start = start_num,
    end = end_num
  )

  expect_s3_class(result, "data.frame")
  expect_true(nrow(result) > 0)
})

# Error handling tests (these don't need vcr as they fail before API call)
test_that("get_parameters validates parameter names", {
  expect_error(
    get_parameters(
      parameters = c("hs", "invalid_param"),
      node = 42
    ),
    "Requested parameters do not exists"
  )
})

test_that("get_parameters validates node input", {
  expect_error(
    get_parameters(
      parameters = "hs",
      node = c(42, 43)
    ),
    "only one location a time"
  )
})

test_that("get_parameters validates date range", {
  expect_error(
    get_parameters(
      parameters = "hs",
      node = 42,
      start = as.POSIXct("1994-01-02 00:00:00", tz = "UTC"),
      end = as.POSIXct("1994-01-01 00:00:00", tz = "UTC")
    ),
    "'end' must be after 'start'"
  )
})

# Test with recorded fixtures to ensure consistent behavior
test_that("get_parameters produces expected data structure over time range", {
  vcr::local_cassette("week_of_data")
  result <- get_parameters(
    parameters = c("hs", "tp"),
    node = 42,
    start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
    end = as.POSIXct("1994-01-07 23:00:00", tz = "UTC")
  )

  expect_s3_class(result, "data.frame")
  expect_equal(ncol(result), 3) # time, hs, tp
  # Expect roughly hourly data for a week (168 hours)
  expect_gt(nrow(result), 100)
  expect_lt(nrow(result), 200)

  # Check data types
  expect_s3_class(result$time, "POSIXct")
  expect_type(result$hs, "double")
  expect_type(result$tp, "double")
})

test_that("get_parameters_raw handles HTTP 404 error", {
  mock_resp <- structure(
    list(status_code = 404),
    class = "httr2_response"
  )

  mockery::stub(get_parameters_raw, "httr2::req_perform", mock_resp)
  mockery::stub(get_parameters_raw, "httr2::resp_status", 404)
  mockery::stub(get_parameters_raw, "httr2::resp_status_desc", "Not Found")

  expect_message(
    result <- get_parameters_raw(
      parameter = "hs",
      node = 42,
      start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
      end = as.POSIXct("1994-01-02 00:00:00", tz = "UTC")
    ),
    "HTTP error 404"
  )

  expect_null(result)
})

test_that("get_parameters_raw handles HTTP 500 server error", {
  mock_resp <- structure(
    list(status_code = 500),
    class = "httr2_response"
  )

  mockery::stub(get_parameters_raw, "httr2::req_perform", mock_resp)
  mockery::stub(get_parameters_raw, "httr2::resp_status", 500)
  mockery::stub(
    get_parameters_raw,
    "httr2::resp_status_desc",
    "Internal Server Error"
  )

  expect_message(
    result <- get_parameters_raw(
      parameter = "hs",
      node = 42,
      start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
      end = as.POSIXct("1994-01-02 00:00:00", tz = "UTC")
    ),
    "HTTP error 500"
  )

  expect_null(result)
})

test_that("get_parameters handles network connection failure gracefully", {
  # Mock the internal function to simulate network failure
  mockery::stub(
    get_parameters,
    "get_parameters_raw",
    NULL # Simulates what get_parameters_raw returns on network failure
  )

  expect_message(
    result <- get_parameters(
      parameters = "hs",
      node = 42,
      start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
      end = as.POSIXct("1994-01-02 00:00:00", tz = "UTC")
    ),
    "Failed to retrieve parameter: hs"
  )

  expect_null(result)
})

test_that("get_parameters_raw handles API-level error in response", {
  mock_resp <- structure(
    list(status_code = 200),
    class = "httr2_response"
  )

  mock_json <- list(
    errorcode = 1,
    errormessage = "Invalid node parameter"
  )

  mockery::stub(get_parameters_raw, "httr2::req_perform", mock_resp)
  mockery::stub(get_parameters_raw, "httr2::resp_status", 200)
  mockery::stub(get_parameters_raw, "httr2::resp_body_json", mock_json)

  expect_message(
    result <- get_parameters_raw(
      parameter = "hs",
      node = 42,
      start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
      end = as.POSIXct("1994-01-02 00:00:00", tz = "UTC")
    ),
    "Invalid node parameter"
  )

  expect_null(result)
})

test_that("get_parameters handles failure in get_parameters_raw for single parameter", {
  # Mock get_parameters_raw to return NULL (simulating any failure)
  mockery::stub(get_parameters, "get_parameters_raw", NULL)

  expect_message(
    result <- get_parameters(
      parameters = "hs",
      node = 42,
      start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
      end = as.POSIXct("1994-01-02 00:00:00", tz = "UTC")
    ),
    "Failed to retrieve parameter: hs"
  )

  expect_null(result)
})

test_that("get_parameters handles partial failure with multiple parameters", {
  # First call succeeds, second call fails
  mock_success <- tibble::tibble(
    time = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
    hs = 1.5
  )

  mockery::stub(
    get_parameters,
    "get_parameters_raw",
    mockery::mock(mock_success, NULL, cycle = TRUE)
  )

  expect_message(
    result <- get_parameters(
      parameters = c("hs", "tp"),
      node = 42,
      start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
      end = as.POSIXct("1994-01-02 00:00:00", tz = "UTC")
    ),
    "Failed to retrieve parameter: tp"
  )

  expect_null(result)
})


test_that("Errors in 'get_parameters()' are handled correcly", {
  expect_error(
    get_parameters("tépé"),
    "Requested parameters do not exists in the database: tépé"
  )
  expect_error(
    get_parameters(node = 0),
    "The requested location do no exist in the database."
  )
  expect_error(
    get_parameters(node = c(10, 100)),
    "The function can retreive only one location a time."
  )
  expect_error(
    get_parameters(start = 1),
    paste0(
      "'start' is outside the covered period: ",
      paste(
        format(
          c(
            resourcecode:::rscd_casandra_start_date,
            resourcecode:::rscd_casandra_end_date
          ),
          format = "%Y-%m-%d %H:%M %Z"
        ),
        collapse = " \u2014 "
      )
    )
  )
  expect_error(
    get_parameters(end = 1e10),
    paste0(
      "'end' is outside the covered period: ",
      paste(
        format(
          c(
            resourcecode:::rscd_casandra_start_date,
            resourcecode:::rscd_casandra_end_date
          ),
          format = "%Y-%m-%d %H:%M %Z"
        ),
        collapse = " \u2014 "
      )
    )
  )
  expect_error(
    get_parameters(
      start = "1994-01-31 01:00:00",
      end = "1994-01-11 01:00:00"
    ),
    "'end' must be after 'start'"
  )
})

test_that("get_parameters accepts dates at exact boundaries", {
  vcr::local_cassette("boundary_dates")

  expect_no_error(
    get_parameters(
      parameters = "hs",
      node = 42,
      start = resourcecode:::rscd_casandra_start_date,
      end = resourcecode:::rscd_casandra_start_date + 3600 # 1 hour later
    )
  )
})

test_that("tp parameter conversion handles edge cases", {
  vcr::local_cassette("tp_conversion")

  result <- get_parameters(
    parameters = "tp",
    node = 42,
    start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
    end = as.POSIXct("1994-01-02 00:00:00", tz = "UTC")
  )

  # All tp values should be positive and finite
  expect_true(all(is.finite(result$tp)))
  expect_true(all(result$tp > 0))
})

#' Compute the dispersion relation of waves
#'  Find *k* s.t. (2.pi.f)^2 = g.k.tanh(k.d)
#'
#' @param frequencies frequency vector
#' @param depth depth (m)
#' @param iter_max maximum number of iterations in the solver
#' @param tol tolerance for termination.
#'
#' @return the wave numbers (same size as frequencies)
#' @export
#'
#' @examples
#' freq <- seq(from = 0, to = 1, length.out = 100)
#' k1 <- dispersion(freq, depth = 1)
#' k10 <- dispersion(freq, depth = 10)
#' kInf <- dispersion(freq, depth = Inf)
#' plot(freq, k1, type = "l")
#' lines(freq, k10, col = "red")
#' lines(freq, kInf, col = "green")
dispersion <- function(frequencies, depth, iter_max = 200, tol = 1e-6) {
  g <- 9.81
  infinite_depth_dispersion <- (4 * pi^2 / g) * frequencies^2

  if (is.infinite(depth)) {
    return(infinite_depth_dispersion)
  }

  frequencies <- as.vector(frequencies)
  out <- frequencies

  for (f in seq_along(frequencies)) {
    if (frequencies[f] == 0) {
      out[f] <- 0
    } else {
      c0 <- (2 * pi * frequencies[f])^2
      k0 <- 4.0243 * frequencies[f]^2
      xk <- k0
      ftest <- 99
      for (ii in 1:iter_max) {
        z <- xk * depth
        y <- tanh(z)
        ff <- c0 - g * xk * y
        dff <- g * (z * (y^2 - 1) - y)
        xk_old <- xk
        xk <- xk_old - ff / dff
        ftest <- abs((xk - xk_old) / xk_old)
        if (ftest <= tol) {
          break
        }
      }
      ff <- c0 - g * xk * y
      out[f] <- xk
    }
  }
  out
}

#' Converts a 2D spectrum time series to a 1D spectrum
#'
#' @param spec a structure with the needed fields, as an output from 'get_2Dspectrum' for example
#' @param ... unused yet
#'
#' @return a structure comparable to 'get_1Dspectrum'.
#' @export
#'
#' @examples
#' spec <- resourcecodedata::rscd_2d_spectra
#' spec1D_RSCD <- resourcecodedata::rscd_1d_spectra
#' spec1D <- convert_spectrum_2d1d(spec)
#' # Check the differences, should be low
#' max(abs(spec1D_RSCD$ef - spec1D$ef))
#'
#' # Plot the different spectrum
#' plot(spec1D$freq, spec1D$ef[, 1], type = "l", log = "xy")
#' lines(spec1D_RSCD$freq, spec1D_RSCD$ef[, 1], col = "red")
#'
#' # Images
#' lims <- c(0, 360)
#' r <- as.POSIXct(round(range(spec1D$forcings$time), "hours"))
#' oldpar <- par(mfcol = c(2, 1))
#' image(spec1D$forcings$time, spec1D$freq, t(spec1D$th1m),
#'   zlim = lims,
#'   xlab = "Time",
#'   ylab = "Freq (Hz)",
#'   xaxt = "n",
#'   main = "Directionnal spreading"
#' )
#' axis.POSIXct(1, spec1D$forcings$time,
#'   at = seq(r[1], r[2], by = "week"),
#'   format = "%Y-%m-%d",
#'   las = 2
#' )
#' image(spec1D_RSCD$forcings$time, spec1D_RSCD$freq, t(spec1D_RSCD$th1m),
#'   zlim = lims,
#'   xlab = "Time",
#'   ylab = "Freq (Hz)",
#'   xaxt = "n"
#' )
#' axis.POSIXct(1, spec1D$forcings$time,
#'   at = seq(r[1], r[2], by = "week"),
#'   format = "%Y-%m-%d",
#'   las = 2
#' )
#' par(oldpar)
convert_spectrum_2d1d <- function(spec, ...) {
  ddir <- diff(spec$dir)[1] * pi / 180 # computes the discretization in direction

  # Averages the directional spectrum along the direction
  # Simple sum is preferred to pracma::trapz for consistency with WWIII internals
  spec$ef <- apply(spec$efth * ddir, seq_along(dim(spec$efth))[-1], sum)

  # Computes the Mean directions and directionnal spreadings
  # (see 2.229 to 2.236 of WWIII user manual for the definitions

  # c1 = cos(theta)*F(theta,f)
  c1 <- sweep(spec$efth, 1, as.array(cos(spec$dir * pi / 180)), FUN = "*")
  # s1 = sin*F(theta,f)
  s1 <- sweep(spec$efth, 1, as.array(sin(spec$dir * pi / 180)), FUN = "*")
  # c2 = cos(2*theta)*F(theta,f)
  c2 <- sweep(spec$efth, 1, as.array(cos(2 * spec$dir * pi / 180)), FUN = "*")
  # s2 = sin(2*theta)*F(theta,f)
  s2 <- sweep(spec$efth, 1, as.array(sin(2 * spec$dir * pi / 180)), FUN = "*")

  # Same notation as the definitions above
  a1 <- apply(c1 * ddir, c(2, 3), sum)
  b1 <- apply(s1 * ddir, c(2, 3), sum)
  a2 <- apply(c2 * ddir, c(2, 3), sum)
  b2 <- apply(s2 * ddir, c(2, 3), sum)

  spec$th1m <- (atan2(b1, a1) * 180 / pi + 180) %% 360
  spec$th2m <- (atan2(b2, a2) * 180 / pi + 180) %% 360
  spec$sth1m <- sqrt(2 * (1 - sqrt((a1^2 + b1^2) / spec$ef^2))) * 180 / pi
  spec$sth2m <- sqrt(2 * (1 - sqrt((a2^2 + b2^2) / spec$ef^2))) * 180 / pi

  # Removes the 2D directional spectrum
  spec$efth <- NULL
  spec$dir <- NULL

  # In 1D spectra, the current direction is taken as "going to"
  spec$forcings$curdir <- (spec$forcings$curdir + 180) %% 360

  # reorder the elements to match the output from `get_1D_spectra`
  spec[c(1:4, 8:12, 5:7)]
}

#' Compute sea_state parameter from wave directional spectrum
#'
#' @param spec 2D spectrum data, e.g. from `get_2Dspectrum`
#' @param ... currently unused
#'
#' @return a tibble with the sea-state parameters computed from the time series of 2D spectrum
#' @export
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' rscd_params <- get_parameters(
#'   node = "134865",
#'   start = "1994-01-01",
#'   end = "1994-01-31 23:00:00",
#'   parameters = c("hs", "tp")
#' )
#' spec <- resourcecodedata::rscd_2d_spectra
#' param_calc <- compute_sea_state_2d_spectrum(spec)
#' oldpar <- par(mfcol = c(2, 1))
#' plot(param_calc$time, param_calc$hs, type = "l", xlab = "Time", ylab = "Hs (m)")
#' lines(rscd_params$time, rscd_params$hs, col = "red")
#' plot(param_calc$time, param_calc$tp, type = "l", xlab = "Time", ylab = "Tp (s)")
#' lines(rscd_params$time, rscd_params$tp, col = "red")
#' par(oldpar)
compute_sea_state_2d_spectrum <- function(spec, ...) {
  # Define an internal function that will do the job for a time-step
  # spectrum: 1D spectrum
  water_density <- 1026
  g <- 9.81

  # Compute 1d spectrum
  ddir <- diff(spec$dir)[1] * pi / 180
  spec_1d <- apply(spec$efth * ddir, c(2, 3), sum)

  # Compute spectral moments
  m0 <- apply(spec_1d, 2, pracma::trapz, x = spec$freq)
  m1 <- apply(
    sweep(spec_1d, 1, spec$freq, FUN = "*"),
    2,
    pracma::trapz,
    x = spec$freq
  )
  m2 <- apply(
    sweep(spec_1d, 1, spec$freq^2, FUN = "*"),
    2,
    pracma::trapz,
    x = spec$freq
  )
  me <- apply(
    sweep(spec_1d, 1, 1 / spec$freq, FUN = "*"),
    2,
    pracma::trapz,
    x = spec$freq
  )

  out <- tibble::tibble(
    time = spec$forcings$time,
    hs = 4 * sqrt(m0),
    t01 = m0 / m1,
    t02 = sqrt(m0 / m2),
    te = me / m0
  )

  # fp evaluaton using spline fitting around spec_1d peak
  nk <- length(spec$freq)

  # Augment frequency resolution by 30
  freqp <- stats::approx(
    1:nk,
    spec$freq,
    xout = seq(
      from = 1,
      to = nk,
      length = 30 * nk
    )
  )
  spec_1d_smooth <- apply(spec_1d, 2, function(y) {
    stats::spline(
      x = spec$freq,
      xout = freqp$y,
      method = "natural",
      y
    )$y
  }) # natural (i.e. "cubic") spline

  fp <- freqp$y[apply(spec_1d_smooth, 2, which.max)]
  out$tp <- 1 / fp

  # Get the forcings fields
  out$dpt <- spec$forcings$dpt
  out$wnd <- spec$forcings$wnd
  out$wnddir <- spec$forcings$wnddir
  out$cur <- spec$forcings$cur

  # In 1D spectral data, the current is "going to" cf user manual so we have to convert
  out$curdir <- (spec$forcings$curdir + 180) %% 360

  # Spectral Bandwidth and Peakedness parameter (Goda 1970)
  out$nu <- sqrt((m0 * m2) / (m1^2) - 1)
  out$mu <- sqrt(1 - m1^2 / (m0 * m2))

  # wave length
  disper_vec <- Vectorize(dispersion, vectorize.args = c("depth"))
  k <- disper_vec(spec$freq, spec$forcings$dpt, iter_max = 200, tol = 1e-6)
  kd <- k * as.numeric(spec$forcings$dpt)

  out$km <- apply(k * spec_1d, 2, pracma::trapz, x = spec$freq) / m0
  out$lm <- 2 * pi / out$km

  # Group velocity
  c1 <- 1 + 2 * kd / sinh(2 * kd)
  c2 <- sqrt(g * tanh(kd) / k)
  cg <- 0.5 * c1 * c2

  # Energy flux
  out$cge <- water_density *
    g *
    apply(cg * spec_1d, 2, pracma::trapz, x = spec$freq) /
    1000
  # convert to kW/m to be consistent with outputs from WWIII

  # Compute mean direction from  and spreading (°)
  aa <- sweep(spec$efth, 1, as.array(cos(spec$dir * pi / 180)), FUN = "*")
  bb <- sweep(spec$efth, 1, as.array(sin(spec$dir * pi / 180)), FUN = "*")

  am <- apply(apply(aa * ddir, c(2, 3), sum), 2, pracma::trapz, x = spec$freq)
  bm <- apply(apply(bb * ddir, c(2, 3), sum), 2, pracma::trapz, x = spec$freq)

  out$dir <- (atan2(bm, am) * 180 / pi + 180) %% 360
  out$spr <- (sqrt(2 * (1 - sqrt((am^2 + bm^2) / m0^2))) * 180 / pi) %% 360

  # Compute mean direction and spreading at peak frequency (°)
  ind_efm <- apply(spec_1d, 2, which.max) # peak of the spectrum
  efm <- array(0, dim = c(36, dim(spec$efth)[3]))

  for (t in seq_len(dim(spec$efth)[3])) {
    efm[, t] <- spec$efth[, ind_efm[t], t]
  } # don't know how to avoid the loop, maybe using slice.index ?

  apm <- apply(
    sweep(efm, 1, as.array(cos(spec$dir * pi / 180)), "*") * ddir,
    2,
    sum
  )
  bpm <- apply(
    sweep(efm, 1, as.array(sin(spec$dir * pi / 180)), "*") * ddir,
    2,
    sum
  )

  out$dp <- (atan2(bpm, apm) * 180 / pi + 180) %% 360

  qpf <- apply(sweep(spec$efth^2, 1, spec$freq, FUN = "*") * ddir, c(2, 3), sum)
  mq <- apply(qpf, 2, pracma::trapz, x = spec$freq)
  out$qp <- ((2 * mq / (m0^2)) * 180 / pi) %% 360

  out
}

#' Compute sea_state parameter from wave spectrum
#'
#' @param spec 1D spectrum data, e.g. from `get_1Dspectrum`
#' @param ... currently unused
#'
#' @return a tibble with the sea-state parameters computed from the time series of 1D spectrum
#' @export
#'
#' @examplesIf identical(tolower(Sys.getenv("NOT_CRAN")), "true")
#' rscd_params <- get_parameters(
#'   node = "134865",
#'   start = "1994-01-01",
#'   end = "1994-01-31 23:00:00",
#'   parameters = c("hs", "tp")
#' )
#' spec <- resourcecodedata::rscd_1d_spectra
#' param_calc <- compute_sea_state_1d_spectrum(spec)
#' oldpar <- par(mfcol = c(2, 1))
#' plot(param_calc$time, param_calc$hs, type = "l", xlab = "Time", ylab = "Hs (m)")
#' lines(rscd_params$time, rscd_params$hs, col = "red")
#' plot(param_calc$time, param_calc$tp, type = "l", xlab = "Time", ylab = "Tp (s)")
#' lines(rscd_params$time, rscd_params$tp, col = "red")
#' par(oldpar)
compute_sea_state_1d_spectrum <- function(spec, ...) {
  water_density <- 1026
  g <- 9.81

  # Compute spectral moments
  m0 <- apply(spec$ef, 2, pracma::trapz, x = spec$freq)
  m1 <- apply(
    sweep(spec$ef, 1, spec$freq, FUN = "*"),
    2,
    pracma::trapz,
    x = spec$freq
  )
  m2 <- apply(
    sweep(spec$ef, 1, spec$freq^2, FUN = "*"),
    2,
    pracma::trapz,
    x = spec$freq
  )
  me <- apply(
    sweep(spec$ef, 1, 1 / spec$freq, FUN = "*"),
    2,
    pracma::trapz,
    x = spec$freq
  )

  out <- tibble::tibble(
    time = spec$forcings$time,
    hs = 4 * sqrt(m0),
    t01 = m0 / m1,
    t02 = sqrt(m0 / m2),
    te = me / m0
  )

  # fp evaluaton using spline fitting around Ef peak
  nk <- length(spec$freq)
  # Augment frequency resolution by 30
  freqp <- stats::approx(
    1:nk,
    spec$freq,
    xout = seq(
      from = 1,
      to = nk,
      length = 30 * nk
    )
  )
  spec_1d_smooth <- apply(spec$ef, 2, function(y) {
    stats::spline(
      x = spec$freq,
      xout = freqp$y,
      method = "natural",
      y
    )$y
  }) # natural (i.e. "cubic") spline

  fp <- freqp$y[apply(spec_1d_smooth, 2, which.max)]
  out$tp <- 1 / fp

  # Get the forcings fields
  out$dpt <- spec$forcings$dpt
  out$wnd <- spec$forcings$wnd
  out$wnddir <- spec$forcings$wnddir
  out$cur <- spec$forcings$cur
  out$curdir <- spec$forcings$curdir

  # Spectral Bandwidth and Peakedness parameter (Goda 1970)
  out$nu <- sqrt((m0 * m2) / (m1^2) - 1)
  out$mu <- sqrt(1 - m1^2 / (m0 * m2))

  # wave length
  disper_vec <- Vectorize(dispersion, vectorize.args = c("depth"))
  k <- disper_vec(spec$freq, spec$forcings$dpt, iter_max = 200, tol = 1e-6)
  kd <- k * as.numeric(spec$forcings$dpt)

  out$km <- apply(k * spec$ef, 2, pracma::trapz, x = spec$freq) / m0
  out$lm <- 2 * pi / out$km

  # Group velocity
  c1 <- 1 + 2 * kd / sinh(2 * kd)
  c2 <- sqrt(g * tanh(kd) / k)
  cg <- 0.5 * c1 * c2

  # Energy flux, converted to kW/m to be consistent with outputs from WWIII
  out$cge <- water_density *
    g *
    apply(cg * spec$ef, 2, pracma::trapz, x = spec$freq) /
    1000

  # Compute mean direction from  and spreading (°)

  out$dir <- apply(spec$th1m, 2, pracma::trapz, x = spec$freq) # possibly wrong
  out$spr <- apply(spec$sth1m, 2, pracma::trapz, x = spec$freq) # possibly wrong

  # mean direction at peak frequency (°)
  ind_efm <- apply(spec$ef, 2, which.max) # peak of the spectrum
  dp <- vector("numeric", dim(spec$ef)[2])

  for (t in seq_len(dim(spec$ef)[2])) {
    dp[t] <- spec$th1m[ind_efm[t], t]
  } # dont know how to avoid the loop, maybe using slice.index?
  out$dp <- dp

  out
}

#' Value Matching
#'
#' @param x value to search
#' @param table table of values
#'
#' @return the opposite of x %in% table
#' @export
#'
#' @examples
#' 1:10 %in% c(1, 3, 5, 9)
#' 1:10 %nin% c(1, 3, 5, 9)
"%nin%" <- function(x, table) {
  !(x %in% table)
}


#' Find the closest point of the FIELD grid to the specified position
#'
#' @param x vector of coordinates in the form longitude/latitude data frame
#' @param lat alternatively, x and lat can be vector of the same length
#' @param closest an integer to specify the number of point to output.
#' @param ... currently unused
#'
#' @return a list with two components: the closest point(s) of the grid and the distance (s).
#' @export
#'
#' @examples
#' semrev_west <- closest_point_field(c(-2.786, 47.239))
#' semrev_west
closest_point_field <- function(x, lat = NULL, closest = 1L, ...) {
  if (!is.null(lat)) {
    if (!(length(x) == length(lat))) {
      stop("If 'lat' is provided, it sould be the same length as 'x'.")
    }
    return(closest_point_field(
      cbind(x, lat),
      lat = NULL,
      closest = closest,
      ...
    ))
  }

  # Concert the input to a matrix with 2 columns, one point per line
  x <- matrix(x, ncol = 2)

  if (!is.numeric(closest) || closest < 1L) {
    stop("'closest' must be an integer greater than 1.")
  }

  # Compute the distance matrix between the points and Resourcecode's grid
  dist <- geosphere::distm(x, resourcecodedata::rscd_field[, c(2, 3)])

  if (closest == 1L) {
    ind_min <- apply(dist, 1, which.min)
  } else {
    ind_min <- apply(dist, 1, function(distances) {
      order(distances)[1:closest]
    })
  }

  distance <- matrix(NA, NROW(x), closest)
  for (p in seq_len(NROW(x))) {
    distance[p, ] <- if (closest == 1) {
      dist[p, ind_min[p]]
    } else {
      dist[p, ind_min[, p]]
    }
  }

  list(
    points = t(ind_min),
    distances = distance
  )
}

#' Find the closest point of the SPEC grid to the specified position
#'
#' @param x vector of coordinates in the form longitude/latitude data frame
#' @param lat alternatively, x and lat can be vector of the same length
#' @param closest an integer to specify the number of point to output.
#' @param ... currently unused
#'
#' @return a list with two components: the closest point(s) of the grid and the distance (s).
#' @export
#'
#' @examples
#' semrev_west <- closest_point_spec(c(-2.786, 47.239))
#' semrev_west
closest_point_spec <- function(x, lat = NULL, closest = 1L, ...) {
  if (!is.null(lat)) {
    if (!(length(x) == length(lat))) {
      stop("If 'lat' is provided, it sould be the same length as 'x'.")
    }
    return(closest_point_spec(
      cbind(x, lat),
      lat = NULL,
      closest = closest,
      ...
    ))
  }

  # Concert the input to a matrix with 2 columns, one point per line
  x <- matrix(x, ncol = 2)

  if (!is.numeric(closest) || closest < 1L) {
    stop("'closest' must be an integer greater than 1.")
  }

  # Compute the distance matrix between the points and Resourcecode's grid
  dist <- geosphere::distm(x, resourcecodedata::rscd_spectral[, c(1, 2)])

  if (closest == 1L) {
    ind_min <- apply(dist, 1, which.min)
  } else {
    ind_min <- apply(dist, 1, function(distances) {
      order(distances)[1:closest]
    })
  }

  distance <- matrix(NA, NROW(x), closest)
  for (p in seq_len(NROW(x))) {
    distance[p, ] <- if (closest == 1) {
      dist[p, ind_min[p]]
    } else {
      dist[p, ind_min[, p]]
    }
  }

  list(
    points = t(ind_min),
    distances = distance
  )
}

#' Convert u/v to meteorological wind speed and direction
#'
#' Converts wind or current zonal and meridional velocity components to
#' magnitude and direction according to meteorological convention.
#'
#' @param u zonal velocity (1D vector) or matrix with zonal and meridional velocity (Nx2 matrix)
#' @param v meridional velocity (1D vector)
#' @param names names to construct the resulting data.frame
#'
#' @return a Nx2 data.frame with the norm and direction (meteorological convention)
#' @export
#'
#' @examples
#' u <- matrix(rnorm(200), nrow = 100, ncol = 2)
#' vdir <- zmcomp2metconv(u)
zmcomp2metconv <- function(u, v = NULL, names = c("wspd", "wdir")) {
  if (is.vector(u)) {
    stopifnot(length(v) == length(u))
    u <- cbind(u, v)
  }

  stopifnot(is.matrix(u) & dim(u)[2] == 2)

  speed <- sqrt(u[, 1]^2 + u[, 2]^2)
  direction <- (270 - atan2(u[, 2], u[, 1]) * 180 / pi) %% 360
  out <- data.frame(speed, direction)
  names(out) <- names
  out
}

#' Convert meteorological wind speed and direction to u/v components
#'
#' @description
#' Converts wind speed (magnitude) and direction (in degrees, meteorological
#' convention: direction from which the wind blows, measured clockwise from north)
#' into zonal (u) and meridional (v) components.
#'
#' @param speed Numeric vector of wind speeds.
#' @param direction Numeric vector of wind directions in degrees (0° = from north,
#' 90° = from east, 180° = from south, 270° = from west).
#' @param names (optional) ames to construct the resulting data.frame.
#'
#' @return A data.frame with two columns:
#' \describe{
#'   \item{u}{Zonal wind component (m/s), positive eastward.}
#'   \item{v}{Meridional wind component (m/s), positive northward.}
#' }
#'
#' @examples
#' # Example 1: North wind of 10 m/s (blowing southward)
#' metconv2zmcomp(10, 0)
#'
#' # Example 2: East wind of 5 m/s (blowing westward)
#' metconv2zmcomp(5, 90)
#'
#' # Example 3: South wind of 8 m/s (blowing northward)
#' metconv2zmcomp(8, 180)
#'
#' @export
metconv2zmcomp <- function(speed, direction, names = c("uwnd", "vwnd")) {
  # Convert to radians
  dir_rad <- direction * pi / 180

  # Components
  u <- -speed * sin(dir_rad)
  v <- -speed * cos(dir_rad)

  out <- data.frame(u = u, v = v)
  names(out) <- names
  out
}


#' JONWSAP spectrum
#'
#' Creates a JONWSAP density spectrum (one-sided), defined by its integral parameters.
#'
#' Reference :
#'   - O.G.Houmb and T.Overvik, "Parametrization of Wave Spectra and Long Term
#'        Joint Distribution of Wave Height and Period,"
#'        in Proceedings, First International Conference
#'        on Behaviour of Offshore Structures (BOSS), Trondheim 1976.
#'        23rd International Towing Tank Conference, vol. II, pp. 544-551
#'   - ITTC Committee, 2002, "The Specialist Committee on Waves -
#'        Final Report and Recommendations to the 23rd ITTC",
#'        Proc. ITTC, vol. II, pp. 505-736.
#'
#' @param hs Hs (default: 5m)
#' @param tp Period (default: 10s)
#' @param fmax higher frequency of the spectrum or
#'             vector of frequencies (default to resourcecode frequency vector)
#' @param df frequency step (unused if fmax=vector of frequencies)
#' @param gam peak enhancement factor (default: 3.3)
#'
#' @return Density spectrum with corresponding parameters
#' @export
#'
#' @examples
#' S1 <- jonswap(tp = 15)
#' S2 <- jonswap(tp = 15, fmax = 0.95, df = 0.003)
#' plot(S1, type = "l", ylim = c(0, 72))
#' lines(S2, col = "red")
#' abline(v = 1 / 15)
jonswap <- function(
  hs = 5,
  tp = 15,
  fmax = resourcecodedata::rscd_freq,
  df = NULL,
  gam = 3.3
) {
  if (length(fmax) > 1) {
    # Case when the frequency vector if given
    freq <- fmax
    fmin <- min(freq)
    fmax <- max(freq)
    df <- min(diff(freq))
    # generate a uniform-sampling to ease computations:
    frq <- seq(from = fmin, to = fmax + df, by = df)
  } else {
    if (is.null(df)) {
      stop("df must be provided when fmax is a single value")
    }
    fmin <- df
    frq <- freq <- seq(from = df, to = fmax + df, by = df)
  }
  nptsp <- length(frq)

  # Stops if the Gamma parameter is lower than 1
  if (gam < 1) {
    stop("Gamma parameter `gam` should be greater than 1")
  }

  # Compute the parameters of the spectrum
  fm <- 1 / tp
  lgam <- log(gam)
  fr <- frq * tp
  ifm <- trunc(fm / df)

  # sigma = c(.07*ones(1,ifm),.09*ones(1,nptsp-ifm))
  # sigma=rep(c(0.07,0.09),c(ifm,nptsp-ifm))
  sigma <- c(0.07 * rep(1, ifm), 0.09 * rep(1, nptsp - ifm))
  frm4 <- fr^-4
  sp <- (1 - fr) / sigma
  sp <- exp(-0.5 * (sp^2))
  sp <- (frm4 / fr) * exp(-1.25 * frm4) * exp(lgam * sp)

  sp <- sp * (hs / 4)^2 / sum(sp * df)
  sp <- Re(c(0, sp))
  sp <- sp[1:nptsp]
  fr <- c(0, frq)
  fr <- fr[1:nptsp]

  sp <- stats::approx(frq, sp, freq)
  names(sp) <- c("freq", "spec")
  attr(sp, "Note") <- paste0(
    "JONSWAP Spectrum, Hs=",
    hs,
    ", Tp=",
    tp,
    ", gamma=",
    gam
  )
  sp$freq <- as.vector(sp$freq)
  tibble::as_tibble(sp)
}

#' Mean Direction
#'
#' Function for computing the (weighted) arithmetic mean of
#' directional data in \strong{meteorological convention}.
#'
#' @param directions numeric vector of directions, in degree, 0° being the North
#' @param weights numeric vector, usually wind speed of wave height.
#'
#' @returns
#' The (weighted) mean of the values in \code{directions} is computed.
#'
#' @export
#'
#' @examples
#' # Test with some wind directions (unweighted)
#' wind_directions <- c(10, 20, 350, 5, 15) # Directions mostly around North
#' mean_dir <- mean_direction(wind_directions)
#' cat("Mean wind direction (unweighted):", round(mean_dir, 1), "degrees\n")
# Test with wind speeds as weights
#' wind_directions <- c(350, 10, 20, 340, 30) # Directions around North
#' wind_speeds <- c(15, 5, 2, 12, 3) # Higher speeds for directions closer to North
#' mean_dir_weighted <- mean_direction(wind_directions, wind_speeds)
#' cat("Mean wind direction (weighted):", round(mean_dir_weighted, 1), "degrees\n")
#'
#' # Compare weighted vs unweighted for the same data
#' mean_dir_unweighted <- mean_direction(wind_directions)
#' cat("Same data unweighted:", round(mean_dir_unweighted, 1), "degrees\n")
mean_direction <- function(directions, weights = NULL) {
  # Verify that directions are numeric
  if (!is.numeric(directions)) {
    stop("'directions' must be numeric")
  }
  # If weights provided, check they have the same length as directions
  if (!is.null(weights)) {
    if (length(directions) != length(weights)) {
      stop("Length of 'directions' and 'speeds' must be equal")
    }
    # Verify that weights are numeric
    if (!is.numeric(weights)) {
      stop("'weights' must be numeric")
    }
    valid_indices <- !is.na(directions) & !is.na(weights)
    directions <- directions[valid_indices]
    weights <- weights[valid_indices]
  } else {
    # Remove NA values from directions only
    directions <- directions[!is.na(directions)]
  }

  # Check for negative weights (which would cause issues with weighting)
  if (any(weights < 0)) {
    warning("Negative weights detected. Using absolute values.")
    weights <- abs(weights)
  }

  # Ensure directions are in [0, 360) range
  directions <- directions %% 360

  # Check if we have any valid directions
  if (length(directions) == 0) {
    return(numeric(0))
  }

  # Set default weights (equal weighting) if weights not provided
  if (is.null(weights)) {
    weights <- rep(1, length(directions))
  }

  # Convert degrees to radians
  radians <- directions * pi / 180

  # Convert to unit vectors (x = sin, y = cos for meteorological convention)
  # In meteorological convention: North = 0°, East = 90°
  x_components <- sin(radians)
  y_components <- cos(radians)

  # Calculate weighted mean components
  mean_x <- stats::weighted.mean(x_components, w = weights)
  mean_y <- stats::weighted.mean(y_components, w = weights)

  # Calculate mean direction in radians
  mean_radians <- atan2(mean_x, mean_y)

  # Convert back to degrees
  mean_degrees <- mean_radians * 180 / pi

  # Ensure result is in [0, 360) range using modulo
  mean_degrees <- mean_degrees %% 360

  mean_degrees
}

#' Directional binning
#'
#' Cuts direction vector into directional bins
# with North bin always centred on 0 degrees.
#'
#' @param directions vector of directions to be binned, in degree, 0° being the North.
#' @param n_bins number of bins, default: 8 sectors.
#' @param labels optional character vector giving the sectors names.
#'
#' @returns
#' a factor vector the same size as \code{directions} with the values binned into sectors.
#'
#' @export
#'
#' @examples
#' # Example usage and demonstration
#' set.seed(123)
#' directions <- runif(20, 0, 360)
#'
#' # Test with different numbers of bins
#' cat("Original directions:\n")
#' print(round(directions, 1))
#'
#' cat("\n8 bins (default):\n")
#' bins_8 <- cut_directions(directions, n_bins = 8)
#' print(bins_8)
#'
#' cat("\n4 bins:\n")
#' bins_4 <- cut_directions(directions, n_bins = 4)
#' print(bins_4)
cut_directions <- function(directions, n_bins = 8, labels = NULL) {
  # Validate inputs
  if (!is.numeric(directions)) {
    stop("'directions' must be numeric")
  }

  if (length(directions) == 0) {
    return(numeric(0))
  }

  if (n_bins < 2) {
    stop("'n_bins' must be at least 2")
  }

  # Normalize directions to 0-360 range
  directions <- directions %% 360

  # Calculate bin width
  bin_width <- 360 / n_bins
  half_bin <- bin_width / 2

  # Create breaks - North bin is centered on 0
  # So breaks go from -half_bin to 360-half_bin
  breaks <- seq(-half_bin, 360 - half_bin, by = bin_width)

  # Adjust directions for the split North bin
  # Values in the upper part of North bin (360-half_bin to 360)
  # need to be mapped to negative values (-half_bin to 0)
  adjusted_directions <- ifelse(
    directions > (360 - half_bin),
    directions - 360,
    directions
  )

  # Create labels if not provided
  if (is.null(labels)) {
    # Calculate center angles for each bin
    centers <- seq(0, 360 - bin_width, by = bin_width)

    # Create descriptive labels based on number of bins
    if (n_bins == 4) {
      labels <- c("N", "E", "S", "W")
    } else if (n_bins == 8) {
      labels <- c("N", "NE", "E", "SE", "S", "SW", "W", "NW")
    } else if (n_bins == 16) {
      labels <- c(
        "N",
        "NNE",
        "NE",
        "ENE",
        "E",
        "ESE",
        "SE",
        "SSE",
        "S",
        "SSW",
        "SW",
        "WSW",
        "W",
        "WNW",
        "NW",
        "NNW"
      )
    } else {
      # Generic labels with center angles
      labels <- paste0("Bin_", sprintf("%.0f", centers))
    }
  }
  # Cut the adjusted directions
  result <- cut(
    adjusted_directions,
    breaks = breaks,
    labels = labels,
    include.lowest = TRUE,
    right = FALSE
  )

  result
}


#' Get season from date time object
#'
#' @param datetime a POSIXct vector from with the season is constructed
#' @param definition the definition used to compute the season. See details section.
#' @param hemisphere in the Southern hemisphere, seasons are reversed compared to the Northern one.
#' @param labels optional, a character vector of length fours with the seasons' names.
#'
#' @returns a Factor vector with 4 levels depending on the definitions (and labels if provided)
#' @export
#'
#' @details
#' Available Definitions:
#'   - meteorological: Standard seasons (Dec-Feb = Winter, etc.)
#'   - astronomical: Based on equinoxes/solstices
#'   - djf: Dec-Jan-Feb, Mar-Apr-May, Jun-Jul-Aug, Sep-Oct-Nov
#'   - jfm: Jan-Feb-Mar, Apr-May-Jun, Jul-Aug-Sep, Oct-Nov-Dec
#'   - fma: Feb-Mar-Apr, May-Jun-Jul, Aug-Sep-Oct, Nov-Dec-Jan
#'   - amj, jas, ond: Alternative starting points for quarterly seasons
#'
#' @examples
#' dates <- seq(
#'   from = as.POSIXct("2023-01-15"),
#'   to = as.POSIXct("2023-12-15"),
#'   by = "month"
#' )
#' cut_seasons(dates)
cut_seasons <- function(
  datetime,
  definition = "meteorological",
  hemisphere = "northern",
  labels = NULL
) {
  # Validate inputs
  if (!inherits(datetime, "POSIXct")) {
    stop("datetime must be a POSIXct object")
  }

  if (
    !definition %in%
      c(
        "meteorological",
        "astronomical",
        "djf",
        "jfm",
        "amj",
        "jas",
        "ond",
        "fma"
      )
  ) {
    stop(
      "definition must be one of: 'meteorological', 'astronomical',
         'djf', 'jfm', 'amj', 'jas', 'ond', 'fma'"
    )
  }

  if (!hemisphere %in% c("northern", "southern")) {
    stop("hemisphere must be 'northern' or 'southern'")
  }

  # Extract month and day-of-year
  month <- as.numeric(format(datetime, "%m"))
  yday <- as.numeric(format(datetime, "%j"))
  year <- as.numeric(format(datetime, "%Y"))

  # Define season boundaries based on definition
  if (definition == "meteorological") {
    # Dec-Jan-Feb = Winter, Mar-Apr-May = Spring, etc.
    season_month <- c(12, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)
    season_labels <- c(
      "Winter",
      "Winter",
      "Winter",
      "Spring",
      "Spring",
      "Spring",
      "Summer",
      "Summer",
      "Summer",
      "Autumn",
      "Autumn",
      "Autumn"
    )
    seasons <- season_labels[match(month, season_month)]
  } else if (definition == "astronomical") {
    # Based on equinoxes and solstices (approximate dates)
    # Northern hemisphere: Dec 21 - Mar 20 = Winter, etc.
    seasons <- rep(NA, length(datetime))

    for (i in seq_along(datetime)) {
      y <- year[i]
      d <- yday[i]

      # Approximate astronomical dates (can vary by 1-2 days)
      spring_equinox <- as.numeric(format(
        as.Date(paste(y, "03", "20", sep = "-")),
        "%j"
      ))
      summer_solstice <- as.numeric(format(
        as.Date(paste(y, "06", "21", sep = "-")),
        "%j"
      ))
      autumn_equinox <- as.numeric(format(
        as.Date(paste(y, "09", "22", sep = "-")),
        "%j"
      ))
      winter_solstice <- as.numeric(format(
        as.Date(paste(y, "12", "21", sep = "-")),
        "%j"
      ))

      if (d >= winter_solstice || d < spring_equinox) {
        seasons[i] <- "Winter"
      } else if (d >= spring_equinox && d < summer_solstice) {
        seasons[i] <- "Spring"
      } else if (d >= summer_solstice && d < autumn_equinox) {
        seasons[i] <- "Summer"
      } else {
        seasons[i] <- "Autumn"
      }
    }
  } else if (definition == "djf") {
    # Dec-Jan-Feb, Mar-Apr-May, Jun-Jul-Aug, Sep-Oct-Nov
    season_month <- c(12, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11)
    season_labels <- c(
      "DJF",
      "DJF",
      "DJF",
      "MAM",
      "MAM",
      "MAM",
      "JJA",
      "JJA",
      "JJA",
      "SON",
      "SON",
      "SON"
    )
    seasons <- season_labels[match(month, season_month)]
  } else if (definition == "jfm") {
    # Jan-Feb-Mar, Apr-May-Jun, Jul-Aug-Sep, Oct-Nov-Dec
    season_month <- 1:12
    season_labels <- c(
      "JFM",
      "JFM",
      "JFM",
      "AMJ",
      "AMJ",
      "AMJ",
      "JAS",
      "JAS",
      "JAS",
      "OND",
      "OND",
      "OND"
    )
    seasons <- season_labels[match(month, season_month)]
  } else if (definition == "amj") {
    # Apr-May-Jun, Jul-Aug-Sep, Oct-Nov-Dec, Jan-Feb-Mar
    season_month <- 1:12
    season_labels <- c(
      "JFM",
      "JFM",
      "JFM",
      "AMJ",
      "AMJ",
      "AMJ",
      "JAS",
      "JAS",
      "JAS",
      "OND",
      "OND",
      "OND"
    )
    seasons <- season_labels[match(month, season_month)]
  } else if (definition == "jas") {
    # Jul-Aug-Sep, Oct-Nov-Dec, Jan-Feb-Mar, Apr-May-Jun
    season_month <- 1:12
    season_labels <- c(
      "JFM",
      "JFM",
      "JFM",
      "AMJ",
      "AMJ",
      "AMJ",
      "JAS",
      "JAS",
      "JAS",
      "OND",
      "OND",
      "OND"
    )
    seasons <- season_labels[match(month, season_month)]
  } else if (definition == "ond") {
    # Oct-Nov-Dec, Jan-Feb-Mar, Apr-May-Jun, Jul-Aug-Sep
    season_month <- 1:12
    season_labels <- c(
      "JFM",
      "JFM",
      "JFM",
      "AMJ",
      "AMJ",
      "AMJ",
      "JAS",
      "JAS",
      "JAS",
      "OND",
      "OND",
      "OND"
    )
    seasons <- season_labels[match(month, season_month)]
  } else if (definition == "fma") {
    # Feb-Mar-Apr, May-Jun-Jul, Aug-Sep-Oct, Nov-Dec-Jan
    season_month <- 1:12
    season_labels <- c(
      "NDJ",
      "FMA",
      "FMA",
      "FMA",
      "MJJ",
      "MJJ",
      "MJJ",
      "ASO",
      "ASO",
      "ASO",
      "NDJ",
      "NDJ"
    )
    seasons <- season_labels[match(month, season_month)]
  }

  # Flip seasons for southern hemisphere
  if (
    hemisphere == "southern" &&
      definition %in% c("meteorological", "astronomical")
  ) {
    season_mapping <- c(
      "Spring" = "Autumn",
      "Summer" = "Winter",
      "Autumn" = "Spring",
      "Winter" = "Summer"
    )
    seasons <- season_mapping[seasons]
  }

  # Define level order BEFORE applying custom labels
  if (definition == "meteorological" || definition == "astronomical") {
    if (hemisphere == "northern") {
      level_order <- c("Spring", "Summer", "Autumn", "Winter")
    } else {
      level_order <- c("Autumn", "Winter", "Spring", "Summer")
    }
  } else if (definition == "djf") {
    level_order <- c("DJF", "MAM", "JJA", "SON")
  } else if (definition == "jfm") {
    level_order <- c("JFM", "AMJ", "JAS", "OND")
  } else if (definition == "fma") {
    level_order <- c("FMA", "MJJ", "ASO", "NDJ")
  } else {
    level_order <- c("JFM", "AMJ", "JAS", "OND") # Default for amj, jas, ond
  }

  # Apply custom labels if provided
  if (!is.null(labels)) {
    unique_seasons <- unique(seasons[!is.na(seasons)])
    if (length(labels) != length(unique_seasons)) {
      stop(paste(
        "Number of labels (",
        length(labels),
        ") must match number of unique seasons (",
        length(unique_seasons),
        ")",
        sep = ""
      ))
    }
    # Create mapping from old to new labels (based on level_order)
    label_mapping <- stats::setNames(labels, level_order)
    seasons <- label_mapping[seasons]
    # Update level order to use custom labels
    level_order <- labels
  }

  result <- factor(seasons, levels = level_order)
  result
}

#' Compute Fractional Day of Year from POSIXct
#'
#' Calculates the fractional day of year from a POSIXct datetime object.
#' The fractional day is zero-indexed, starting at 0 for January 1st at
#' midnight and ending at approximately 365.958 for December 31st at 23:00
#' in a non-leap year (or 366.958 in a leap year).
#'
#' @param datetime A POSIXct object or vector of POSIXct objects representing
#'   date-time values. Must have a timezone attribute.
#'
#' @return A numeric vector of the same length as \code{datetime}, containing
#'   fractional day of year values. January 1st at midnight corresponds to 0,
#'   and each hour adds approximately 0.04167 (1/24) to the value.
#'
#' @details
#' The function computes the time difference in hours between the input
#' datetime and midnight on January 1st of the same year, then divides by 24
#' to obtain fractional days. The calculation accounts for leap years
#' automatically.
#'
#' @examples
#' dates <- seq(
#'   from = as.POSIXct("2024-01-01 00:00:00", tz = "UTC"),
#'   to = as.POSIXct("2024-01-02 00:00:00", tz = "UTC"),
#'   by = "6 hours"
#' )
#' fractional_day_of_year(dates) # Returns 0.00, 0.25, 0.50, 0.75, 1.00
#'
#' # End of year
#' dt_end <- as.POSIXct("2024-12-31 23:00:00", tz = "UTC")
#' fractional_day_of_year(dt_end) # Returns ~365.958
#'
#' @export
fractional_day_of_year <- function(datetime) {
  # Input validation
  if (!inherits(datetime, "POSIXct")) {
    stop("datetime must be a POSIXct object")
  }

  # Extract timezone (defaults to UTC if missing)
  tz <- attr(datetime, "tzone")
  if (is.null(tz) || tz == "") {
    tz <- "UTC"
  }

  # Get the year for each datetime
  year <- as.POSIXlt(datetime, tz = tz)$year + 1900

  # Initialize output
  fractional_day <- rep(NA_real_, length(datetime))

  # Process only non-NA entries
  valid <- !is.na(datetime)
  if (any(valid)) {
    # Start of the year in same timezone
    start_of_year <- as.POSIXct(
      paste0(year[valid], "-01-01 00:00:00"),
      tz = tz
    )

    # Compute time difference in hours
    hours_diff <- as.numeric(difftime(
      datetime[valid],
      start_of_year,
      units = "hours"
    ))

    # Convert to fractional days
    fractional_day[valid] <- hours_diff / 24
  }

  fractional_day
}

test_that("fractional_day_of_year returns 0 for January 1st at midnight", {
  dt <- as.POSIXct("2024-01-01 00:00:00", tz = "UTC")
  expect_equal(fractional_day_of_year(dt), 0)

  # Test for different year
  dt2 <- as.POSIXct("2023-01-01 00:00:00", tz = "UTC")
  expect_equal(fractional_day_of_year(dt2), 0)
})

test_that("fractional_day_of_year returns correct value for December 31st at 23:00", {
  # Non-leap year (2023)
  dt_2023 <- as.POSIXct("2023-12-31 23:00:00", tz = "UTC")
  expect_equal(fractional_day_of_year(dt_2023), 364 + 23 / 24, tolerance = 1e-6)

  # Leap year (2024)
  dt_2024 <- as.POSIXct("2024-12-31 23:00:00", tz = "UTC")
  expect_equal(fractional_day_of_year(dt_2024), 365 + 23 / 24, tolerance = 1e-6)
})

test_that("fractional_day_of_year correctly handles hourly increments", {
  dates <- seq(
    from = as.POSIXct("2024-01-01 00:00:00", tz = "UTC"),
    to = as.POSIXct("2024-01-01 23:00:00", tz = "UTC"),
    by = "hour"
  )

  result <- fractional_day_of_year(dates)
  expected <- seq(0, 23 / 24, by = 1 / 24)

  expect_equal(result, expected, tolerance = 1e-10)
  expect_length(result, 24)
})

test_that("fractional_day_of_year correctly handles noon", {
  dt <- as.POSIXct("2024-01-01 12:00:00", tz = "UTC")
  expect_equal(fractional_day_of_year(dt), 0.5)

  # Mid-year
  dt_midyear <- as.POSIXct("2024-07-01 12:00:00", tz = "UTC")
  # July 1st is day 183 in a leap year (182 complete days have passed)
  expect_equal(fractional_day_of_year(dt_midyear), 182.5, tolerance = 1e-6)
})

test_that("fractional_day_of_year handles leap years correctly", {
  # February 29th in leap year
  dt_leap <- as.POSIXct("2024-02-29 00:00:00", tz = "UTC")
  # Feb 29 is the 60th day (59 complete days have passed)
  expect_equal(fractional_day_of_year(dt_leap), 59)

  # March 1st in leap year vs non-leap year
  dt_leap_mar <- as.POSIXct("2024-03-01 00:00:00", tz = "UTC")
  dt_nonleap_mar <- as.POSIXct("2023-03-01 00:00:00", tz = "UTC")

  expect_equal(fractional_day_of_year(dt_leap_mar), 60) # Day 61 in leap year
  expect_equal(fractional_day_of_year(dt_nonleap_mar), 59) # Day 60 in non-leap year
})

test_that("fractional_day_of_year works with vectors", {
  dates <- c(
    as.POSIXct("2024-01-01 00:00:00", tz = "UTC"),
    as.POSIXct("2024-01-01 06:00:00", tz = "UTC"),
    as.POSIXct("2024-01-01 12:00:00", tz = "UTC"),
    as.POSIXct("2024-01-01 18:00:00", tz = "UTC")
  )

  result <- fractional_day_of_year(dates)
  expected <- c(0, 0.25, 0.5, 0.75)

  expect_equal(result, expected)
  expect_length(result, 4)
})

test_that("fractional_day_of_year handles different timezones", {
  dt_utc <- as.POSIXct("2024-03-07 12:00:00", tz = "UTC")
  dt_est <- as.POSIXct("2024-03-07 12:00:00", tz = "America/New_York")

  # Both should calculate based on their respective midnight Jan 1
  result_utc <- fractional_day_of_year(dt_utc)
  result_est <- fractional_day_of_year(dt_est)

  # They should give the same fractional day for the same local time
  expect_equal(result_utc, result_est)

  # June 15 is day 167 in a leap year (166 complete days)
  expect_equal(result_utc, 66.5, tolerance = 1e-6)
})

test_that("fractional_day_of_year errors with non-POSIXct input", {
  expect_error(
    fractional_day_of_year("2024-01-01"),
    "datetime must be a POSIXct object"
  )

  expect_error(
    fractional_day_of_year(as.Date("2024-01-01")),
    "datetime must be a POSIXct object"
  )

  expect_error(
    fractional_day_of_year(123456),
    "datetime must be a POSIXct object"
  )
})

test_that("fractional_day_of_year handles edge cases across year boundary", {
  # Last hour of year
  dt_last <- as.POSIXct("2024-12-31 23:00:00", tz = "UTC")
  result_last <- fractional_day_of_year(dt_last)

  # First hour of next year
  dt_first <- as.POSIXct("2025-01-01 00:00:00", tz = "UTC")
  result_first <- fractional_day_of_year(dt_first)

  expect_equal(result_first, 0)
  expect_gt(result_last, 365) # Should be greater than 365
})

test_that("fractional_day_of_year is consistent with sequential dates", {
  dates <- seq(
    from = as.POSIXct("2024-01-01 00:00:00", tz = "UTC"),
    to = as.POSIXct("2024-01-10 00:00:00", tz = "UTC"),
    by = "day"
  )

  result <- fractional_day_of_year(dates)

  # Should increase by 1 each day
  diffs <- diff(result)
  expect_true(all(abs(diffs - 1) < 1e-10))

  # First value should be 0, last should be 9
  expect_equal(result[1], 0)
  expect_equal(result[length(result)], 9)
})

test_that("fractional_day_of_year handles NA values", {
  dates <- c(
    as.POSIXct("2024-01-01 00:00:00", tz = "UTC"),
    as.POSIXct(NA, tz = "UTC"),
    as.POSIXct("2024-01-01 12:00:00", tz = "UTC")
  )

  result <- fractional_day_of_year(dates)

  expect_equal(result[1], 0)
  expect_true(is.na(result[2]))
  expect_equal(result[3], 0.5)
})

# ==============================================================================
# Test Suite 1: Input Validation
# ==============================================================================

test_that("Input validation works correctly", {
  # Test ill-formed coordinates input
  expect_error(
    closest_point_field(c(1, 2), 1),
    "If 'lat' is provided, it sould be the same length as 'x'."
  )

  expect_error(
    closest_point_field(c(1, 2), closest = "one"),
    "'closest' must be an integer greater than 1."
  )
  expect_error(
    closest_point_field(c(1, 2), closest = 0),
    "'closest' must be an integer greater than 1."
  )

  # Test ill-formed coordinates input
  expect_error(
    closest_point_spec(c(1, 2), 1),
    "If 'lat' is provided, it sould be the same length as 'x'."
  )

  expect_error(
    closest_point_spec(c(1, 2), closest = "one"),
    "'closest' must be an integer greater than 1."
  )
  expect_error(
    closest_point_spec(c(1, 2), closest = 0),
    "'closest' must be an integer greater than 1."
  )
})

# ==============================================================================
# Test Suite 2: Node selection: one location provided
# ==============================================================================

test_that("Unique node selection works", {
  skip_if(!requireNamespace("resourcecodedata", quietly = TRUE))

  point_test <- c(-4.700, 48.302)
  node_test <- matrix(134938)
  spec_test <- matrix(2124)

  expect_equal(closest_point_field(point_test)$point, node_test)
  expect_equal(
    closest_point_field(point_test[1], point_test[2])$point,
    node_test
  )
  expect_equal(closest_point_spec(point_test)$point, spec_test)
  expect_equal(
    closest_point_spec(point_test[1], point_test[2])$point,
    spec_test
  )
})

# ==============================================================================
# Test Suite 3: Multiple nodes selection
# ==============================================================================

test_that("Multiple nodes selection works", {
  skip_if(!requireNamespace("resourcecodedata", quietly = TRUE))

  point_test <- matrix(
    c(
      resourcecodedata::rscd_field[c(100, 1000), 2],
      resourcecodedata::rscd_field[c(100, 1000), 3]
    ),
    ncol = 2
  )
  node_test <- matrix(c(100, 1000), ncol = 2)
  spec_test <- matrix(c(23045, 22274), ncol = 2)

  output <- closest_point_field(point_test)
  expect_named(output, c("points", "distances"))
  expect_identical(dim(output$points), c(1L, 2L))
  expect_identical(dim(output$distances), c(2L, 1L))
  expect_equal(output$point, node_test)
  expect_true(all(output$distance == 0))

  output2 <- closest_point_field(point_test[, 1], point_test[, 2])
  expect_equal(output2, output)

  output_spec <- closest_point_spec(point_test)
  expect_named(output_spec, c("points", "distances"))
  expect_identical(dim(output_spec$points), c(1L, 2L))
  expect_identical(dim(output_spec$distances), c(2L, 1L))
  expect_equal(output_spec$point, spec_test)

  output2_spec <- closest_point_spec(point_test[, 1], point_test[, 2])
  expect_equal(output2_spec, output_spec)
})

# ==============================================================================
# Test Suite 4: Several closest points selection
# ==============================================================================

test_that("List of closest nodes selection works", {
  skip_if(!requireNamespace("resourcecodedata", quietly = TRUE))

  point_test <- matrix(
    c(
      resourcecodedata::rscd_field[c(100, 1000), 2],
      resourcecodedata::rscd_field[c(100, 1000), 3]
    ),
    ncol = 2
  )
  node_test <- matrix(c(100, 101, 99, 1000, 1001, 999), ncol = 3, byrow = TRUE)
  closest <- 3

  output <- closest_point_field(point_test, closest = closest)
  expect_named(output, c("points", "distances"))
  expect_identical(dim(output$points), c(2L, 3L))
  expect_identical(dim(output$distances), c(2L, 3L))
  expect_equal(output$point, node_test)
  expect_true(all(output$distance[, 1] == 0))

  output2 <- closest_point_field(
    point_test[, 1],
    point_test[, 2],
    closest = closest
  )
  expect_equal(output2, output)

  output_spec <- closest_point_spec(point_test, closest = 3)
  expect_named(output_spec, c("points", "distances"))
  expect_identical(dim(output_spec$points), c(2L, 3L))
  expect_identical(dim(output_spec$distances), c(2L, 3L))

  output2_spec <- closest_point_spec(
    point_test[, 1],
    point_test[, 2],
    closest = 3
  )
  expect_equal(output2_spec, output_spec)
})

# Test basic functionality
test_that("cut_seasons returns a factor", {
  dates <- seq(
    from = as.POSIXct("2023-01-15", tz = "UTC"),
    to = as.POSIXct("2023-12-15", tz = "UTC"),
    by = "month"
  )

  result <- cut_seasons(dates)

  expect_s3_class(result, "factor")
})

# Test input validation
test_that("cut_seasons rejects non-POSIXct input", {
  dates <- as.Date(c("2023-01-15", "2023-06-15"))

  expect_error(
    cut_seasons(dates),
    "datetime must be a POSIXct object"
  )
})

test_that("cut_seasons rejects invalid definition", {
  dates <- as.POSIXct("2023-01-15", tz = "UTC")

  expect_error(
    cut_seasons(dates, definition = "invalid"),
    "definition must be one of"
  )
})

test_that("cut_seasons rejects invalid hemisphere", {
  dates <- as.POSIXct("2023-01-15", tz = "UTC")

  expect_error(
    cut_seasons(dates, hemisphere = "eastern"),
    "hemisphere must be 'northern' or 'southern'"
  )
})

# Test meteorological definition
test_that("cut_seasons meteorological definition works correctly for northern hemisphere", {
  dates <- as.POSIXct(
    c(
      "2023-01-15",
      "2023-03-15",
      "2023-06-15",
      "2023-09-15",
      "2023-12-15"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(
    dates,
    definition = "meteorological",
    hemisphere = "northern"
  )

  expect_equal(
    as.character(result),
    c("Winter", "Spring", "Summer", "Autumn", "Winter")
  )
  expect_equal(levels(result), c("Spring", "Summer", "Autumn", "Winter"))
})

test_that("cut_seasons meteorological definition works correctly for southern hemisphere", {
  dates <- as.POSIXct(
    c(
      "2023-01-15",
      "2023-03-15",
      "2023-06-15",
      "2023-09-15",
      "2023-12-15"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(
    dates,
    definition = "meteorological",
    hemisphere = "southern"
  )

  expect_equal(
    as.character(result),
    c("Summer", "Autumn", "Winter", "Spring", "Summer")
  )
  expect_equal(levels(result), c("Autumn", "Winter", "Spring", "Summer"))
})

# Test astronomical definition
test_that("cut_seasons astronomical definition works correctly", {
  dates <- as.POSIXct(
    c(
      "2023-01-15",
      "2023-04-15",
      "2023-07-15",
      "2023-10-15",
      "2023-12-25"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(
    dates,
    definition = "astronomical",
    hemisphere = "northern"
  )

  expect_equal(
    as.character(result),
    c("Winter", "Spring", "Summer", "Autumn", "Winter")
  )
})

test_that("cut_seasons astronomical definition handles equinox/solstice boundaries", {
  dates <- as.POSIXct(
    c(
      "2023-03-20",
      "2023-06-21",
      "2023-09-22",
      "2023-12-21"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(
    dates,
    definition = "astronomical",
    hemisphere = "northern"
  )

  # These dates should be on or near the boundaries
  expect_true(all(result %in% c("Spring", "Summer", "Autumn", "Winter")))
})

# Test DJF definition
test_that("cut_seasons djf definition works correctly", {
  dates <- as.POSIXct(
    c(
      "2023-01-15",
      "2023-04-15",
      "2023-07-15",
      "2023-10-15",
      "2023-12-15"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(dates, definition = "djf")

  expect_equal(as.character(result), c("DJF", "MAM", "JJA", "SON", "DJF"))
  expect_equal(levels(result), c("DJF", "MAM", "JJA", "SON"))
})

# Test JFM definition
test_that("cut_seasons jfm definition works correctly", {
  dates <- as.POSIXct(
    c(
      "2023-02-15",
      "2023-05-15",
      "2023-08-15",
      "2023-11-15"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(dates, definition = "jfm")

  expect_equal(as.character(result), c("JFM", "AMJ", "JAS", "OND"))
  expect_equal(levels(result), c("JFM", "AMJ", "JAS", "OND"))
})

# Test FMA definition
test_that("cut_seasons fma definition works correctly", {
  dates <- as.POSIXct(
    c(
      "2023-01-15",
      "2023-03-15",
      "2023-06-15",
      "2023-09-15",
      "2023-12-15"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(dates, definition = "fma")

  expect_equal(as.character(result), c("NDJ", "FMA", "MJJ", "ASO", "NDJ"))
  expect_equal(levels(result), c("FMA", "MJJ", "ASO", "NDJ"))
})

# Test AMJ definition
test_that("cut_seasons amj definition works correctly", {
  dates <- as.POSIXct(
    c(
      "2023-01-15",
      "2023-05-15",
      "2023-08-15",
      "2023-11-15"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(dates, definition = "amj")

  expect_equal(as.character(result), c("JFM", "AMJ", "JAS", "OND"))
})

# Test JAS definition
test_that("cut_seasons jas definition works correctly", {
  dates <- as.POSIXct(
    c(
      "2023-02-15",
      "2023-05-15",
      "2023-08-15",
      "2023-11-15"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(dates, definition = "jas")

  expect_equal(as.character(result), c("JFM", "AMJ", "JAS", "OND"))
})

# Test OND definition
test_that("cut_seasons ond definition works correctly", {
  dates <- as.POSIXct(
    c(
      "2023-02-15",
      "2023-05-15",
      "2023-08-15",
      "2023-11-15"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(dates, definition = "ond")

  expect_equal(as.character(result), c("JFM", "AMJ", "JAS", "OND"))
})

# Test custom labels
test_that("cut_seasons accepts custom labels", {
  dates <- as.POSIXct(
    c(
      "2023-01-15",
      "2023-04-15",
      "2023-07-15",
      "2023-10-15"
    ),
    tz = "UTC"
  )

  custom_labels <- c("Q1", "Q2", "Q3", "Q4")
  result <- cut_seasons(dates, definition = "djf", labels = custom_labels)

  expect_equal(levels(result), custom_labels)
})

test_that("cut_seasons rejects incorrect number of custom labels", {
  dates <- as.POSIXct("2023-01-15", tz = "UTC")

  expect_error(
    cut_seasons(dates, definition = "meteorological", labels = c("A", "B")),
    "Number of labels"
  )
})

# Test edge cases
test_that("cut_seasons handles single date", {
  date <- as.POSIXct("2023-06-15", tz = "UTC")

  result <- cut_seasons(date)

  expect_length(result, 1)
  expect_equal(as.character(result), "Summer")
})

test_that("cut_seasons handles year boundaries", {
  dates <- as.POSIXct(
    c(
      "2023-12-31 23:00:00",
      "2024-01-01 00:00:00"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(dates, definition = "meteorological")

  expect_equal(as.character(result), c("Winter", "Winter"))
})

test_that("cut_seasons handles leap years", {
  dates <- as.POSIXct(
    c(
      "2024-02-29 12:00:00",
      "2023-03-01 12:00:00"
    ),
    tz = "UTC"
  )

  result <- cut_seasons(dates, definition = "meteorological")

  expect_equal(as.character(result), c("Winter", "Spring"))
})

test_that("cut_seasons handles month boundaries correctly", {
  # Test boundaries between months for meteorological definition
  dates <- as.POSIXct(
    c(
      "2023-02-28 23:59:59", # Last second of February (Winter)
      "2023-03-01 00:00:00" # First second of March (Spring)
    ),
    tz = "UTC"
  )

  result <- cut_seasons(dates, definition = "meteorological")

  expect_equal(as.character(result), c("Winter", "Spring"))
})

test_that("cut_seasons handles all 12 months correctly", {
  dates <- as.POSIXct(paste0("2023-", 1:12, "-15"), tz = "UTC")

  result <- cut_seasons(
    dates,
    definition = "meteorological",
    hemisphere = "northern"
  )

  expected <- c(
    "Winter",
    "Winter",
    "Spring",
    "Spring",
    "Spring",
    "Summer",
    "Summer",
    "Summer",
    "Autumn",
    "Autumn",
    "Autumn",
    "Winter"
  )

  expect_equal(as.character(result), expected)
})

test_that("cut_seasons preserves vector length", {
  dates <- seq(
    from = as.POSIXct("2023-01-01", tz = "UTC"),
    to = as.POSIXct("2023-12-31", tz = "UTC"),
    by = "day"
  )

  result <- cut_seasons(dates)

  expect_length(result, length(dates))
})

test_that("cut_seasons works with different timezones", {
  dates_utc <- as.POSIXct("2023-06-15 12:00:00", tz = "UTC")
  dates_est <- as.POSIXct("2023-06-15 12:00:00", tz = "America/New_York")

  result_utc <- cut_seasons(dates_utc)
  result_est <- cut_seasons(dates_est)

  # Both should classify as Summer (same month)
  expect_equal(as.character(result_utc), "Summer")
  expect_equal(as.character(result_est), "Summer")
})

test_that("cut_seasons astronomical definition handles year transitions", {
  dates <- as.POSIXct(
    c(
      "2023-12-20 12:00:00", # Before winter solstice
      "2023-12-22 12:00:00" # After winter solstice
    ),
    tz = "UTC"
  )

  result <- cut_seasons(dates, definition = "astronomical")

  expect_equal(as.character(result), c("Autumn", "Winter"))
})

test_that("cut_seasons custom labels work with meteorological definition", {
  dates <- as.POSIXct(
    c(
      "2023-01-15",
      "2023-04-15",
      "2023-07-15",
      "2023-10-15"
    ),
    tz = "UTC"
  )

  custom_labels <- c("Printemps", "Été", "Automne", "Hiver")
  result <- cut_seasons(
    dates,
    definition = "meteorological",
    labels = custom_labels
  )

  expect_true(all(levels(result) %in% custom_labels))
})

test_that("cut_seasons handles vector with NA values", {
  dates <- as.POSIXct(c("2023-01-15", NA, "2023-07-15"), tz = "UTC")

  result <- cut_seasons(dates)

  expect_length(result, 3)
  expect_true(is.na(result[2]))
  expect_equal(as.character(result[c(1, 3)]), c("Winter", "Summer"))
})

#' Helper function to download netCDF file given its URL
#'
#' @param url the remote file to download
#' @param destfile the destination file
#'
#' @returns the path to the destination file, or NULL if an error occurred.
#'
#' @noRd
#' @keywords internal
download_nc <- function(url, destfile) {
  tryCatch(
    {
      req <- httr2::request(url) |>
        httr2::req_error(is_error = \(resp) FALSE) |> # Don't auto-error on HTTP errors
        httr2::req_retry(max_tries = 3) |> # Retry transient failures
        httr2::req_timeout(60) |>
        httr2::req_user_agent("Resourcecode R package") |>
        httr2::req_error(is_error = ~ .x$status_code >= 400)

      httr2::req_perform(req, path = destfile)

      if (!file.exists(destfile)) {
        return(NULL)
      }

      destfile
    },
    error = function(e) {
      message("Download failed: ", conditionMessage(e))
      if (file.exists(destfile)) {
        file.remove(destfile)
      }
      NULL
    }
  )
}


#' Download the 2D spectrum data from IFREMER ftp for a single data.
#'
#' No consistency checks are made, this function should not be called directly by the user
#'
#' @param point the location name (string) requested.
#' @param year the year (as a string) requested.
#' @param month the month number, as a string,
#'
#' @return a list with the sea-state spectrum and forcings
#' @noRd
#' @keywords internal
get_2d_spectrum_raw <- function(point, year, month) {
  base <- "https://data-dataref.ifremer.fr/ww3/resourcecode/HINDCAST/"
  url <- paste0(
    base,
    year,
    "/",
    month,
    "/SPEC_NC/RSCD_WW3-RSCD-UG-",
    point,
    "_",
    year,
    month,
    "_spec.nc"
  )

  temp <- tempfile(fileext = ".nc")

  file <- download_nc(url, temp)

  if (is.null(file)) {
    message(
      "Could not download spectral data.
              The remote server may be unavailable or the URL may have changed."
    )
    return(NULL)
  }

  nc <- ncdf4::nc_open(file)

  on.exit({
    ncdf4::nc_close(nc)
    file.remove(temp)
  })

  freq <- nc$dim$frequency$vals
  dir <- nc$dim$direction$vals
  dir_ordred <- order(dir)

  time <- nc$dim$time$vals
  time <- as.POSIXct(
    time * 24 * 3600,
    origin = "1990-01-01Z00:00:00",
    tz = "UTC"
  )

  var_out <- list("longitude", "latitude", "frequency1", "frequency2", "efth")
  out <- lapply(var_out, ncdf4::ncvar_get, nc = nc)
  names(out) <- var_out

  out$longitude <- out$longitude[1]
  out$latitude <- out$latitude[1]
  out$freq <- freq
  out$dir <- dir[dir_ordred]

  var_out_1d <- list("dpt", "wnd", "wnddir", "cur", "curdir")
  forcings <- lapply(var_out_1d, ncdf4::ncvar_get, nc = nc)
  names(forcings) <- var_out_1d

  out$forcings <- tibble::tibble(time = time, tibble::as_tibble(forcings))

  out$efth <- 10^(out$efth) - 1e-12
  out$efth <- out$efth[dir_ordred, , ]

  out$station <- point
  out
}

#' Download the 1D spectrum data from IFREMER ftp for a single data.
#'
#' No consistency checks are made, this function should not be called directly by the user
#'
#' @param point the location name (string) requested.
#' @param year the year (as a string) requested.
#' @param month the month number, as a string,
#'
#' @return a list with the sea-state 1D spectrum and forcings
#' @noRd
#' @keywords internal
get_1d_spectrum_raw <- function(point, year, month) {
  base <- "https://data-dataref.ifremer.fr/ww3/resourcecode/HINDCAST/"
  url <- paste0(
    base,
    year,
    "/",
    month,
    "/FREQ_NC/RSCD_WW3-RSCD-UG-",
    point,
    "_",
    year,
    month,
    "_freq.nc"
  )
  temp <- tempfile(fileext = ".nc")

  file <- download_nc(url, temp)

  if (is.null(file)) {
    message(
      "Could not download spectral data.
              The remote server may be unavailable or the URL may have changed."
    )
    return(NULL)
  }

  nc <- ncdf4::nc_open(file)

  on.exit({
    ncdf4::nc_close(nc)
    file.remove(temp)
  })

  freq <- nc$dim$frequency$vals

  time <- nc$dim$time$vals
  time <- as.POSIXct(
    time * 24 * 3600,
    origin = "1990-01-01Z00:00:00",
    tz = "UTC"
  )

  var_out <- list(
    "longitude",
    "latitude",
    "frequency1",
    "frequency2",
    "ef",
    "th1m",
    "th2m",
    "sth1m",
    "sth2m"
  )
  out <- lapply(var_out, ncdf4::ncvar_get, nc = nc)
  names(out) <- var_out

  out$longitude <- out$longitude[1]
  out$latitude <- out$latitude[1]
  out$freq <- freq

  var_out_1d <- list(
    "dpt",
    "wnd",
    "wnddir",
    "cur",
    "curdir",
    "hs",
    "fp",
    "f02",
    "f0m1",
    "th1p",
    "sth1p",
    "dir",
    "spr"
  )
  forcings <- lapply(var_out_1d, ncdf4::ncvar_get, nc = nc)

  names(forcings) <- var_out_1d

  out$forcings <- tibble::tibble(time = time, tibble::as_tibble(forcings))

  out$station <- point
  out
}

#' Download the 2D spectrum data from IFREMER ftp
#'
#' @param point the location name (string) requested.
#'              Alternatively, the node number. The consistency is checked internally.
#' @param start the starting date (as a string). The consistency is checked internally.
#' @param end the ending date as a string
#'
#' @return A list with 9 elements:
#' \describe{
#'   \item{longitude}{Longitude}
#'   \item{latitude}{Latitude}
#'   \item{frequency1}{Lower frequency}
#'   \item{frequency2}{Upper frequency}
#'   \item{ef}{Surface elevation variance spectral density}
#'   \item{th1m}{Mean direction from first spectral moment}
#'   \item{th2m}{Mean direction from second spectral moment}
#'   \item{sth1m}{Mean directional spreading from first spectral moment}
#'   \item{sth2m}{Mean directional spreading from second spectral moment}
#'   \item{freq}{Central frequency}
#'   \item{dir}{Directionnal bins}
#'   \item{forcings}{A data.frame with 6 variables:
#'    \describe{
#'      \item{time}{Time}
#'      \item{dpt}{Depth, positive downward}
#'      \item{wnd}{Wind intensity, at 10m above sea level}
#'      \item{wnddir}{Wind direction, comes from}
#'      \item{cur}{Current intensity, at the surface}
#'      \item{curdir}{Current direction, going to}
#'    }}
#'   \item{station}{Station name}
#' }
#' @export
#'
#' @examples
#' spec2D <- get_2d_spectrum("SEMREVO", start = "1994-01-01", end = "1994-02-28")
#' if(!is.null(spec2D)){
#'   image(spec2D$dir, spec2D$freq, spec2D$efth[, , 1],
#'     xlab = "Direction (°)",
#'     ylab = "Frequency (Hz"
#'   )
#' }
get_2d_spectrum <- function(point, start = "1994-01-01", end = "1994-02-28") {
  stopifnot(length(point) == 1)

  if (is.numeric(point)) {
    point <- resourcecodedata::rscd_spectral[point, "name"]
  }

  stopifnot(point %in% resourcecodedata::rscd_spectral$name)

  if (is.character(start)) {
    start <- as.POSIXct(start, tz = "UTC")
  }
  if (is.character(end)) {
    end <- as.POSIXct(end, tz = "UTC")
  }

  if (is.numeric(start)) {
    start <- as.POSIXct(
      start,
      tz = "UTC",
      origin = as.POSIXct("1970-01-01 00:00:00", tz = "UTC")
    )
  }
  if (is.numeric(end)) {
    end <- as.POSIXct(
      end,
      tz = "UTC",
      origin = as.POSIXct("1970-01-01 00:00:00", tz = "UTC")
    )
  }

  stopifnot(format(start, "%Y") >= format(rscd_hindcast_start_date, "%Y"))
  stopifnot(format(end, "%Y") <= format(rscd_hindcast_end_date, "%Y"))
  stopifnot(end >= start)

  dates <- seq.POSIXt(from = start, to = end, by = "month")
  years <- format(dates, format = "%Y")
  months <- format(dates, format = "%m")

  out <- get_2d_spectrum_raw(point, years[1], months[1])

  if (is.null(out)) {
    message(
      "Failed to download data for ",
      point,
      " (",
      years[1],
      "-",
      months[1],
      ")"
    )
    return(NULL)
  }

  for (m in seq_along(years[-1])) {
    temp <- get_2d_spectrum_raw(point, years[m + 1], months[m + 1])

    if (is.null(temp)) {
      message(
        "Failed to download data for ",
        point,
        " (",
        years[m + 1],
        "-",
        months[m + 1],
        ")"
      )
      return(NULL)
    }

    out$efth <- abind::abind(out$efth, temp$efth, along = 3)
    out$forcings <- rbind(out$forcings, temp$forcings)
  }

  out
}


#' Download the 1D spectrum data from IFREMER ftp
#'
#' @param point the location name (string) requested.
#'              Alternatively, the node number. The consistency is checked internally.
#' @param start the starting date (as a string). The consistency is checked internally.
#' @param end the ending date as a string
#'
#' @return A list with 12 elements:
#' \describe{
#'   \item{longitude}{Longitude}
#'   \item{latitude}{Latitude}
#'   \item{frequency1}{Lower frequency}
#'   \item{frequency2}{Upper frequency}
#'   \item{ef}{Surface elevation variance spectral density}
#'   \item{th1m}{Mean direction from first spectral moment}
#'   \item{th2m}{Mean direction from second spectral moment}
#'   \item{sth1m}{Mean directional spreading from first spectral moment}
#'   \item{sth2m}{Mean directional spreading from second spectral moment}
#'   \item{freq}{Central frequency}
#'   \item{forcings}{A data.frame with 14 variables:
#'    \describe{
#'      \item{time}{Time}
#'      \item{dpt}{Depth, positive downward}
#'      \item{wnd}{Wind intensity, at 10m above sea level}
#'      \item{wnddir}{Wind direction, comes from}
#'      \item{cur}{Current intensity, at the surface}
#'      \item{curdir}{Current direction, going to}
#'      \item{hs}{Significant wave height}
#'      \item{fp}{Peak wave frequency}
#'      \item{f02}{Mean wave frequency}
#'      \item{f0m1}{Mean wave frequency at spectral moment minus one}
#'      \item{th1p}{Mean wave direction at spectral peak}
#'      \item{sth1p}{Directional spreading at spectral peak}
#'      \item{dir}{Mean wave direction}
#'      \item{spr}{Mean directional spreading}
#'    }}
#'   \item{station}{Station name}
#' }
#' @export
#'
#' @examples
#' spec1D <- get_1d_spectrum("SEMREVO", start = "1994-01-01", end = "1994-02-28")
#' if(!is.null(spec1D)){
#'   r <- as.POSIXct(round(range(spec1D$forcings$time), "month"))
#'   image(spec1D$forcings$time, spec1D$freq, t(spec1D$ef),
#'     xaxt = "n", xlab = "Time",
#'     ylab = "Frequency (Hz)"
#'   )
#'   axis.POSIXct(1, spec1D$forcings$time,
#'     at = seq(r[1], r[2], by = "week"),
#'     format = "%Y-%m-%d", las = 2
#'   )
#' }
get_1d_spectrum <- function(point, start = "1994-01-01", end = "1994-02-28") {
  stopifnot(length(point) == 1)

  if (is.numeric(point)) {
    point <- resourcecodedata::rscd_spectral[point, "name"]
  }

  stopifnot(point %in% resourcecodedata::rscd_spectral$name)

  if (is.character(start)) {
    start <- as.POSIXct(start, tz = "UTC")
  }
  if (is.character(end)) {
    end <- as.POSIXct(end, tz = "UTC")
  }

  if (is.numeric(start)) {
    start <- as.POSIXct(
      start,
      tz = "UTC",
      origin = as.POSIXct("1970-01-01", tz = "UTC")
    )
  }
  if (is.numeric(end)) {
    end <- as.POSIXct(
      end,
      tz = "UTC",
      origin = as.POSIXct("1970-01-01", tz = "UTC")
    )
  }

  stopifnot(format(start, "%Y") >= format(rscd_hindcast_start_date, "%Y"))
  stopifnot(format(end, "%Y") <= format(rscd_hindcast_end_date, "%Y"))
  stopifnot(end >= start)

  dates <- seq.POSIXt(from = start, to = end, by = "month")
  years <- format(dates, format = "%Y")
  months <- format(dates, format = "%m")

  out <- get_1d_spectrum_raw(point, years[1], months[1])

  if (is.null(out)) {
    message(
      "Failed to download data for ",
      point,
      " (",
      years[1],
      "-",
      months[1],
      ")"
    )
    return(NULL)
  }

  for (m in seq_along(years[-1])) {
    temp <- get_1d_spectrum_raw(point, years[m + 1], months[m + 1])

    if (is.null(temp)) {
      message(
        "Failed to download data for ",
        point,
        " (",
        years[m + 1],
        "-",
        months[m + 1],
        ")"
      )
      return(NULL)
    }

    out$ef <- abind::abind(out$ef, temp$ef, along = 2)
    out$th1m <- abind::abind(out$th1m, temp$th1m, along = 2)
    out$th2m <- abind::abind(out$th2m, temp$th2m, along = 2)
    out$sth1m <- abind::abind(out$sth1m, temp$sth1m, along = 2)
    out$sth2m <- abind::abind(out$sth2m, temp$sth2m, along = 2)
    out$forcings <- rbind(out$forcings, temp$forcings)
  }

  out
}

# Tests for get_1d_spectrum()
test_that("get_1d_spectrum retrieves data successfully", {
  skip_if_offline()
  skip_on_cran()

  spec <- get_1d_spectrum(
    "SEMREVO",
    start = "1994-01-01",
    end = "1994-02-28"
  )

  expect_type(spec, "list")
  expect_named(
    spec,
    c(
      "longitude",
      "latitude",
      "frequency1",
      "frequency2",
      "ef",
      "th1m",
      "th2m",
      "sth1m",
      "sth2m",
      "freq",
      "forcings",
      "station"
    )
  )
  expect_s3_class(spec$forcings, "data.frame")
  expect_shape(spec$forcings, dim = c(1416, 14))
  expect_equal(spec$station, "SEMREVO")
})

test_that("get_1d_spectrum handles numeric node input", {
  #vcr::local_cassette("get_1d_spectrum_numeric_node")
  skip_if_offline()
  skip_on_cran()

  spec_by_name <- get_1d_spectrum(
    "SEMREVO",
    start = "1994-01-01",
    end = "1994-01-31"
  )

  # Find the index of SEMREVO in rscd_spectral
  idx <- which(resourcecodedata::rscd_spectral$name == "SEMREVO")

  spec_by_index <- get_1d_spectrum(
    idx,
    start = "1994-01-01",
    end = "1994-01-31"
  )

  expect_equal(spec_by_name$station, spec_by_index$station)
})

test_that("get_1d_spectrum handles character date inputs", {
  #vcr::local_cassette("get_1d_spectrum_character_dates")
  skip_if_offline()
  skip_on_cran()

  spec <- get_1d_spectrum(
    "SEMREVO",
    start = "1994-01-01",
    end = "1994-01-31"
  )

  expect_type(spec, "list")
  expect_shape(spec$forcings, dim = c(744, 14))
})

test_that("get_1d_spectrum handles numeric (UNIX timestamp) date inputs", {
  # vcr::local_cassette("get_1d_spectrum_numeric_dates")
  skip_if_offline()
  skip_on_cran()

  start_unix <- as.numeric(as.POSIXct("1994-01-01", tz = "UTC"))
  end_unix <- as.numeric(as.POSIXct("1994-01-31", tz = "UTC"))

  spec <- get_1d_spectrum(
    "SEMREVO",
    start = start_unix,
    end = end_unix
  )

  expect_type(spec, "list")
  expect_gt(nrow(spec$forcings), 0)
})

test_that("get_1d_spectrum handles multi-month requests", {
  # vcr::local_cassette("get_1d_spectrum_multi_month")
  skip_if_offline()
  skip_on_cran()

  spec <- get_1d_spectrum(
    "SEMREVO",
    start = "1994-01-01",
    end = "1994-03-31"
  )

  expect_type(spec, "list")
  # Should have data for 3 months
  expect_gt(nrow(spec$forcings), 2000) # Rough estimate
})

test_that("get_1d_spectrum validates forcings data structure", {
  # vcr::local_cassette("get_1d_spectrum_forcings_structure")
  skip_if_offline()
  skip_on_cran()

  spec <- get_1d_spectrum(
    "SEMREVO",
    start = "1994-01-01",
    end = "1994-01-31"
  )

  # Check forcings has expected columns
  expected_cols <- c(
    "time",
    "dpt",
    "wnd",
    "wnddir",
    "cur",
    "curdir",
    "hs",
    "fp",
    "f02",
    "f0m1",
    "th1p",
    "sth1p",
    "dir",
    "spr"
  )
  expect_true(all(expected_cols %in% names(spec$forcings)))

  # Check time is POSIXct
  expect_s3_class(spec$forcings$time, "POSIXct")
})

test_that("get_1d_spectrum validates spectral arrays dimensions", {
  # vcr::local_cassette("get_1d_spectrum_array_dimensions")
  skip_if_offline()
  skip_on_cran()

  spec <- get_1d_spectrum(
    "SEMREVO",
    start = "1994-01-01",
    end = "1994-01-31"
  )

  # Check that spectral arrays have correct dimensions
  expect_true(is.matrix(spec$ef))
  expect_equal(nrow(spec$ef), length(spec$freq))
  expect_equal(ncol(spec$ef), nrow(spec$forcings))
})

# Error handling tests - input validation (no network needed)
test_that("get_1d_spectrum validates point input length", {
  expect_error(
    get_1d_spectrum(
      c("SEMREVO", "AUTRE"),
      start = "1994-01-01",
      end = "1994-01-31"
    ),
    "length\\(point\\) == 1"
  )
})

test_that("get_1d_spectrum validates point exists", {
  expect_error(
    get_1d_spectrum(
      "INVALID_POINT",
      start = "1994-01-01",
      end = "1994-01-31"
    ),
    "point %in% resourcecodedata::rscd_spectral\\$name"
  )
})

test_that("get_1d_spectrum validates date range", {
  expect_error(
    get_1d_spectrum(
      "SEMREVO",
      start = "1994-01-31",
      end = "1994-01-01"
    ),
    "end >= start"
  )
})

test_that("get_1d_spectrum validates start date is within coverage", {
  expect_error(
    get_1d_spectrum(
      "SEMREVO",
      start = "1980-01-01", # Before hindcast period
      end = "1994-01-31"
    ),
    "format\\(start, \"%Y\"\\) >= format\\(rscd_hindcast_start_date"
  )
})

test_that("get_1d_spectrum validates end date is within coverage", {
  expect_error(
    get_1d_spectrum(
      "SEMREVO",
      start = "1994-01-01",
      end = "2030-01-01" # After hindcast period
    ),
    "format\\(end, \"%Y\"\\) <= format\\(rscd_hindcast_end_date"
  )
})

# Network failure tests - using mocks
test_that("get_1d_spectrum fails gracefully when first download fails", {
  # Mock the internal raw function to return NULL (simulating download failure)
  mockery::stub(
    get_1d_spectrum,
    "get_1d_spectrum_raw",
    NULL
  )

  # Should fail or return NULL depending on your implementation
  # Update this based on how you handle NULL in get_1d_spectrum
  expect_null(suppressMessages(
    get_1d_spectrum(
      "SEMREVO",
      start = "1994-01-01",
      end = "1994-01-31"
    )
  )
  )
})

# Tests for get_2d_spectrum()
test_that("get_2d_spectrum retrieves data successfully", {
  # vcr::local_cassette("get_2d_spectrum_basic")
  skip_if_offline()
  skip_on_cran()

  spec <- get_2d_spectrum(
    "SEMREVO",
    start = "1994-01-01",
    end = "1994-01-31"
  )

  expect_type(spec, "list")
  expect_named(
    spec,
    c(
      "longitude",
      "latitude",
      "frequency1",
      "frequency2",
      "efth",
      "freq",
      "dir",
      "forcings",
      "station"
    )
  )
  expect_s3_class(spec$forcings, "data.frame")
  expect_shape(spec$forcings, dim = c(744, 6))
  expect_equal(spec$station, "SEMREVO")
})

test_that("get_2d_spectrum handles numeric node input", {
  # vcr::local_cassette("get_2d_spectrum_numeric_node")
  skip_if_offline()
  skip_on_cran()

  spec_by_name <- get_2d_spectrum(
    "SEMREVO",
    start = "1994-01-01",
    end = "1994-01-31"
  )

  idx <- which(resourcecodedata::rscd_spectral$name == "SEMREVO")

  spec_by_index <- get_2d_spectrum(
    idx,
    start = "1994-01-01",
    end = "1994-01-31"
  )

  expect_equal(spec_by_name$station, spec_by_index$station)
})

test_that("get_2d_spectrum handles numeric date inputs", {
  # vcr::local_cassette("get_2d_spectrum_numeric_dates")
  skip_if_offline()
  skip_on_cran()

  start_unix <- as.numeric(as.POSIXct("1994-01-01", tz = "UTC"))
  end_unix <- as.numeric(as.POSIXct("1994-01-31", tz = "UTC"))

  spec <- get_2d_spectrum(
    "SEMREVO",
    start = start_unix,
    end = end_unix
  )

  expect_type(spec, "list")
  expect_gt(nrow(spec$forcings), 0)
})

test_that("get_2d_spectrum handles multi-month requests", {
  # vcr::local_cassette("get_2d_spectrum_multi_month")
  skip_if_offline()
  skip_on_cran()

  spec <- get_2d_spectrum(
    "SEMREVO",
    start = "1994-01-01",
    end = "1994-02-28"
  )

  expect_type(spec, "list")
  expect_shape(spec$forcings, dim = c(1416, 6))

  # Check forcings has expected columns
  expected_cols <- c("time", "dpt", "wnd", "wnddir", "cur", "curdir")
  expect_true(all(expected_cols %in% names(spec$forcings)))

  # Check time is POSIXct
  expect_s3_class(spec$forcings$time, "POSIXct")

  #Spectral data
  expect_true(is.array(spec$efth))
  expect_equal(length(dim(spec$efth)), 3)
  expect_equal(dim(spec$efth)[1], length(spec$dir))
  expect_equal(dim(spec$efth)[2], length(spec$freq))
  expect_equal(dim(spec$efth)[3], nrow(spec$forcings))
})

# Error handling tests - input validation (no network needed)
test_that("get_2d_spectrum validates point input length", {
  expect_error(
    get_2d_spectrum(
      c("SEMREVO", "AUTRE"),
      start = "1994-01-01",
      end = "1994-01-31"
    ),
    "length\\(point\\) == 1"
  )
})

test_that("get_2d_spectrum validates point exists", {
  expect_error(
    get_2d_spectrum(
      "INVALID_POINT",
      start = "1994-01-01",
      end = "1994-01-31"
    ),
    "point %in% resourcecodedata::rscd_spectral\\$name"
  )
})

test_that("get_2d_spectrum validates date range", {
  expect_error(
    get_2d_spectrum(
      "SEMREVO",
      start = "1994-01-31",
      end = "1994-01-01"
    ),
    "end >= start"
  )
})

test_that("get_2d_spectrum validates start date within coverage", {
  expect_error(
    get_2d_spectrum(
      "SEMREVO",
      start = "1980-01-01",
      end = "1994-01-31"
    ),
    "format\\(start, \"%Y\"\\) >= format\\(rscd_hindcast_start_date"
  )
})

test_that("get_2d_spectrum validates end date within coverage", {
  expect_error(
    get_2d_spectrum(
      "SEMREVO",
      start = "1994-01-01",
      end = "2030-01-01"
    ),
    "format\\(end, \"%Y\"\\) <= format\\(rscd_hindcast_end_date"
  )
})

# Network failure tests - using mocks (no vcr needed)
test_that("get_2d_spectrum fails gracefully when first download fails", {
  mockery::stub(
    get_2d_spectrum,
    "get_2d_spectrum_raw",
    NULL
  )

  expect_null(suppressMessages(
    get_2d_spectrum(
      "SEMREVO",
      start = "1994-01-01",
      end = "1994-01-31"
    )
  )
  )
})

# Edge case: Boundary dates
test_that("get_1d_spectrum accepts dates at exact boundaries", {
  skip_if_offline()
  skip_on_cran()
  #vcr::local_cassette("get_1d_spectrum_boundary_dates")

  # Test with start date at exact boundary
  expect_no_error(
    get_1d_spectrum(
      "SEMREVO",
      start = format(resourcecode:::rscd_hindcast_start_date, "%Y-01-01"),
      end = format(resourcecode:::rscd_hindcast_start_date, "%Y-01-31")
    )
  )
})

test_that("get_2d_spectrum accepts dates at exact boundaries", {
  skip_if_offline()
  skip_on_cran()
  # vcr::local_cassette("get_2d_spectrum_boundary_dates")

  # Test with start date at exact boundary
  expect_no_error(
    get_2d_spectrum(
      "SEMREVO",
      start = format(resourcecode:::rscd_hindcast_start_date, "%Y-01-01"),
      end = format(resourcecode:::rscd_hindcast_start_date, "%Y-01-31")
    )
  )
})

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom ggplot2 ggplot aes geom_path coord_sf theme_void geom_polygon scale_fill_distiller
#' @importFrom ggplot2 scale_color_distiller theme margin element_rect element_blank geom_rect
#' @importFrom Rcpp sourceCpp evalCpp
#' @importFrom rlang .data
#' @useDynLib resourcecode, .registration = TRUE
## usethis namespace: end
NULL

#' Resourcecode data extract
#'
#' An extract of the Resourcecode Hindcast database, at node 123456, spanning from
#'  1994-01-01 00:00:00 to 1999-12-31 23:00:00.
#'
#'@format ## A data frame with 87,648 rows and 7 variables:
#' \describe{
#'   \item{time}{POSIXct. Timestamp in UTC (hourly resolution).}
#'   \item{hs}{numeric. Significant wave height (m).}
#'   \item{tp}{numeric. Peak wave period (s).}
#'   \item{dp}{numeric. Mean wave direction (degrees, coming from true North, clockwise).}
#'   \item{uwnd}{numeric. Zonal (east–west) component of the 10-m wind (m/s). Positive eastward.}
#'   \item{vwnd}{numeric. Meridional (north–south) component of the 10-m wind (m/s).
#'   Positive northward.}
#'   \item{dpt}{numeric.Depth (m).}
#' }
#'
#' @source User Manual of the RESOURCECODE database
#' <https://archimer.ifremer.fr/doc/00751/86306/>
"rscd_data_example"

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' Fast implementation of pracma::trapz from the Armadillo C++ library
#'
#' Compute the area of a multivariate function with (matrix) values Y at the points x.
#'
#' @param x x-coordinates of points on the x-axis (vector)
#' @param Y y-coordinates of function values (matrix)
#' @param dim an integer giving the subscripts which the function will be applied over. 1 indicates rows, 2 indicates columns
#' @return a vector with one dimension less than Y
#' @export
#' @examples
#' x = 1:10
#' Y = sin(pi/10*matrix(1:10,ncol=10,nrow=10))
#' fastTrapz(x*pi/10,Y,2)
fastTrapz <- function(x, Y, dim = 1L) {
    .Call(`_resourcecode_fastTrapz`, x, Y, dim)
}

ww_calc_cpp <- function(times, winlen_hours, allow_overlap, tstep_secs) {
    .Call(`_resourcecode_ww_calc_cpp`, times, winlen_hours, allow_overlap, tstep_secs)
}

#' Compute the orbital speed at a given depth from the wave elevation 1D spectra
#'
#' @param spec 1D spectral data: TxM matrix
#' @param freq the M frequencies
#' @param z distance above the floor at which we want the orbital speed (single numeric)
#' @param depth depth time series (vector length T. Recycled if a single value is given)
#' @param output_speeds TRUE if the spectral speed are needed. Otherwise, returns the RMS (default)
#'
#' @return depending on spec, a list with the spectral velocities for each component
#'         (if output_speeds==FALSE) or a data.frame with a time series of horizontal and vertical
#'         components of (spectral) orbital speed.
#' @export
#'
#' @examples
#' # Compute orbital speed for varying Hs
#' S <- t(sapply(1:10, function(h) {
#'   jonswap(h)$spec
#' }))
#' orb_speeds <- compute_orbital_speeds(S, resourcecodedata::rscd_freq, depth = 100, z = 10)
#' plot(1:10, orb_speeds[, 1],
#'   type = "l",
#'   ylim = range(orb_speeds),
#'   xlab = "Hs (m)",
#'   ylab = "Orbital speed RMS (m/s)"
#' )
#' lines(1:10, orb_speeds[, 2], type = "l", col = "red")
compute_orbital_speeds <- function(
  spec,
  freq,
  z = 0,
  depth = Inf,
  output_speeds = FALSE
) {
  # z: distance above sea floor

  dims <- dim(spec)
  n_time <- dims[1]
  n_freq <- dims[2]

  if (length(depth) == 1) {
    depth <- rep(depth, n_time)
  }

  stopifnot(n_freq == length(freq))
  stopifnot(z >= 0)
  stopifnot(z <= depth)

  # Compute k efficiently when depth is discretized
  k <- outer(
    freq,
    unique(depth),
    Vectorize(
      resourcecode::dispersion,
      vectorize.args = c("frequencies", "depth")
    )
  )
  mat_k <- t(k[, match(depth, unique(depth))])

  mat_d <- matrix(depth, nrow = n_time, ncol = n_freq)
  mat_depth <- matrix(z, nrow = n_time, ncol = n_freq)
  mat_freq <- matrix(freq, nrow = n_time, ncol = n_freq, byrow = TRUE)

  csh1 <- cosh(mat_k * mat_depth)
  ssh1 <- sinh(mat_k * mat_depth)
  ssh2 <- sinh(mat_k * mat_d)

  spectral_u_component <- (2 * pi * mat_freq * csh1 / ssh2)^2 * spec

  spectral_v_component <- (2 * pi * mat_freq * ssh1 / ssh2)^2 * spec

  if (output_speeds) {
    out <- array(NA, dim = c(dim(spectral_u_component), 2))
    out[,, 1] <- spectral_u_component
    out[,, 2] <- spectral_v_component
    out
  } else {
    u_rms <- sqrt(2 * resourcecode::fastTrapz(freq, spectral_u_component, 2))
    v_rms <- sqrt(2 * resourcecode::fastTrapz(freq, spectral_v_component, 2))
    cbind(u_rms, v_rms)
  }
}

---
title: "Introduction to `{resourcecode}`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction to `{resourcecode}`}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  out.width = "100%",
  fig.align = "center",
  fig.retina = 3
)
```

The goal of `{resourcecode}` is to provide an easy access to the ResourceCODE hindcast database of sea-states. More information on the database can be found [here](https://resourcecode.ifremer.fr). We will explain in this vignette the way to retrieve data from this database and some the functionalities offered by this package.


Once the packages are installed, we can proceed as usual:

```{r setup}
library(resourcecodedata)
library(resourcecode)
library(ggplot2)
```

## Database configuration: nodes, bathymetry...

There are two grids in the Resource**CODE** hindcast database: the full, dense, **FIELD** grid, which is the grid where the numerical WWIII model is run and where sea-state parameters and only 1D spectra are available. The second grid, **SPEC**, is coarser, but contains the full 2D spectral data at each node.

- **FIELD** grid

The variable `rscd_field` contains the coordinates of the FIELD grid, along with the depth of these points and `d50`, the bottom sediment type. 

```{r}
str(rscd_field)
head(rscd_field)
```

At each location, a large number of sea-state parameters are available, described in `rscd_variables`:

```{r}
str(rscd_variables)
head(rscd_variables)
```

The available location can be plotted on a map using the following code. It can be noticed here that the variables `rscd_coastline` and `rscd_islands` contain the path of the coast line and the islands, respectively.

```{r, fig.retina=3}
lim_lon <- c(-5.25, -4.25)
lim_lat <- c(47.75, 48.75)
field_bzh <- ggplot(rscd_field, aes(x = longitude, y = latitude)) +
  geom_point(size = .1, col = "lightblue") +
  geom_path(data = rscd_coastline, linewidth = .2) +
  geom_path(data = rscd_islands, aes(group = .data$ID), linewidth = .2) +
  coord_sf(xlim = lim_lon, ylim = lim_lat, expand = FALSE, crs = sf::st_crs(4326)) +
  theme_void()
field_bzh
```

- **SPEC** grid

The spectral grid, or **SPEC** grid is coarser, with more than 24,000 nodes where the full 2D spectra are available. The grid can be added to the previous plot to see the differences in the spatial coverage.

```{r}
str(rscd_spectral)
head(rscd_spectral)
```


```{r, fig.retina=3}
field_bzh + geom_point(data = rscd_spectral, col = "orange", size = .1)
```

- Helpers to find points

When one is interested in some location, we also provide helpers to find the closest point in each of the grid, via the `closest_point_FIELD()` and `closest_point_SPEC()` functions.

## Getting data

We have two different sources of data, the FIELD grid, and the SPEC grid. We will cover each use case here.

- Sea-state parameters from the FIELD can be retrieved using the `get_parameters()` function, which allows to download a time series easily at a given location. For the moment (Sept. 2023), the 1D spectra on this grid are not available to direct download.

For example, if one is interested in the time series of sea-state parameters at some location, the following code can be adapted:

```{r, fig.height=4,fig.width=8,warning=FALSE,message=FALSE}
point_of_interest <- c(longitude = -4.6861533, latitude = 48.3026514)
node <- closest_point_field(point_of_interest)
node
ts <- get_parameters(node = node$point, parameters = c("hs", "tp", "dp", "cge"))
if (!is.null(ts)) {
  ggplot(tidyr::pivot_longer(ts, -1), aes(x = time, y = value, col = name)) +
    geom_line() +
    coord_cartesian(expand = FALSE) +
    facet_wrap(~name, ncol = 2, scales = "free_y") +
    scale_x_datetime(name = NULL, date_breaks = "month") +
    scale_y_continuous(name = NULL) +
    theme_minimal() +
    theme(
      legend.position = "none",
      axis.text.x = element_text(angle = 60, hjust = 1)
    )
}
```

- 1D and 2D spectra of the SPEC grid can be downloaded directly from IFREMER FTP using functions `get_1Dspectrum()` and `get_2Dspectrum()`. We also provide a plotting function for the 2D spectrum.

```{r}
node_spectral_grid <- closest_point_spec(point_of_interest)
```

Then one can extract the 1D spectrum time-series and see the contents of these downloaded data.

```{r, eval = FALSE}
spec_1d <- get_1d_spectrum(node_spectral_grid$point, start = "1994-01-01", end = "1994-02-28")
str(spec_1d)
```

We also provide in the package `{resourcecodedata}` pre-loaded time series of 1D spectra, mainly for testing and example purpose:


```{r, eval = FALSE}
spec_1d <- resourcecodedata::rscd_1d_spectra
str(spec_1d)
```


The same applies to 2D spectra:

```{r, eval = FALSE}
spec_2d <- get_2d_spectrum(node_spectral_grid$point, start = "1994-01-01", end = "1994-02-28")
str(resourcecodedata::rscd_2d_spectra)
```


In addition, it is possible to plot the wave elevation directional spectra for any given time, which can be specified by the time index or directly the date:

```{r,warning=FALSE, fig.height=8,fig.width=8}
plot_2d_specta(resourcecodedata::rscd_2d_spectra, "1994-01-15 18:00")
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/download_data.R
\name{get_parameters}
\alias{get_parameters}
\title{Download time series of sea-state parameters from RESOURCECODE database}
\usage{
get_parameters(
  parameters = "hs",
  node = 42,
  start = as.POSIXct("1994-01-01 00:00:00", tz = "UTC"),
  end = as.POSIXct("1994-12-31 23:00:00", tz = "UTC")
)
}
\arguments{
\item{parameters}{character vector of sea-state parameters}

\item{node}{single integer with the node to get}

\item{start}{starting date (as integer, character or posixct)}

\item{end}{ending date (as integer, character or posixct)}
}
\value{
a tibble with N-rows and \code{length(parameters)} columns.
}
\description{
If the remote resource is unavailable or returns an error, the function returns NULL
and emits an informative message.
}
\examples{
rscd_data <- get_parameters(parameters = c("hs", "tp"), node = 42)
if(!is.null(rscd_data)) plot(rscd_data$time, rscd_data$hs, type = "l")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/specta_data_ploting.R
\name{plot_1d_specta}
\alias{plot_1d_specta}
\title{Plot a wave density 1D spectrum at a given time}
\usage{
plot_1d_specta(spec, time = 1L, print_sea_state = TRUE, ...)
}
\arguments{
\item{spec}{the spectral data, as an output from \code{get_2Dspectrum}}

\item{time}{the time to plot. Either an integer or the date.}

\item{print_sea_state}{should the sea_states parameters being plot ? Default to TRUE.}

\item{...}{currently unused}
}
\value{
a ggplot object
}
\description{
Plot a wave density 1D spectrum at a given time
}
\examples{
plot_1d_specta(resourcecodedata::rscd_1d_spectra, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{closest_point_field}
\alias{closest_point_field}
\title{Find the closest point of the FIELD grid to the specified position}
\usage{
closest_point_field(x, lat = NULL, closest = 1L, ...)
}
\arguments{
\item{x}{vector of coordinates in the form longitude/latitude data frame}

\item{lat}{alternatively, x and lat can be vector of the same length}

\item{closest}{an integer to specify the number of point to output.}

\item{...}{currently unused}
}
\value{
a list with two components: the closest point(s) of the grid and the distance (s).
}
\description{
Find the closest point of the FIELD grid to the specified position
}
\examples{
semrev_west <- closest_point_field(c(-2.786, 47.239))
semrev_west
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{rscd_data_example}
\alias{rscd_data_example}
\title{Resourcecode data extract}
\format{
\subsection{A data frame with 87,648 rows and 7 variables:}{

\describe{
\item{time}{POSIXct. Timestamp in UTC (hourly resolution).}
\item{hs}{numeric. Significant wave height (m).}
\item{tp}{numeric. Peak wave period (s).}
\item{dp}{numeric. Mean wave direction (degrees, coming from true North, clockwise).}
\item{uwnd}{numeric. Zonal (east–west) component of the 10-m wind (m/s). Positive eastward.}
\item{vwnd}{numeric. Meridional (north–south) component of the 10-m wind (m/s).
Positive northward.}
\item{dpt}{numeric.Depth (m).}
}
}
}
\source{
User Manual of the RESOURCECODE database
\url{https://archimer.ifremer.fr/doc/00751/86306/}
}
\usage{
rscd_data_example
}
\description{
An extract of the Resourcecode Hindcast database, at node 123456, spanning from
1994-01-01 00:00:00 to 1999-12-31 23:00:00.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{fastTrapz}
\alias{fastTrapz}
\title{Fast implementation of pracma::trapz from the Armadillo C++ library}
\usage{
fastTrapz(x, Y, dim = 1L)
}
\arguments{
\item{x}{x-coordinates of points on the x-axis (vector)}

\item{Y}{y-coordinates of function values (matrix)}

\item{dim}{an integer giving the subscripts which the function will be applied over. 1 indicates rows, 2 indicates columns}
}
\value{
a vector with one dimension less than Y
}
\description{
Compute the area of a multivariate function with (matrix) values Y at the points x.
}
\examples{
x = 1:10
Y = sin(pi/10*matrix(1:10,ncol=10,nrow=10))
fastTrapz(x*pi/10,Y,2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resourcecode-package.R
\docType{package}
\name{resourcecode-package}
\alias{resourcecode}
\alias{resourcecode-package}
\title{resourcecode: Access to the 'RESOURCECODE' Hindcast Database}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Utility functions to download data from the 'RESOURCECODE' hindcast database of sea-states, time series of sea-state parameters and time series of 1D and 2D wave spectra. See \url{https://resourcecode.ifremer.fr} for more details about the available data. Also provides facilities to plot and analyse downloaded data, such as computing the sea-state parameters from both the 1D and 2D surface elevation variance spectral density.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/Resourcecode-project/r-resourcecode}
  \item \url{https://resourcecode-project.r-universe.dev/resourcecode}
  \item Report bugs at \url{https://github.com/Resourcecode-project/r-resourcecode/issues}
}

}
\author{
\strong{Maintainer}: Nicolas Raillard \email{nicolas.raillard@ifremer.fr} (\href{https://orcid.org/0000-0003-3385-5104}{ORCID})

Authors:
\itemize{
  \item Nicolas Raillard \email{nicolas.raillard@ifremer.fr} (\href{https://orcid.org/0000-0003-3385-5104}{ORCID})
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/specta_data_ploting.R
\name{plot_2d_specta}
\alias{plot_2d_specta}
\title{Plot a wave density 2D spectrum}
\usage{
plot_2d_specta(
  spec,
  time = 1L,
  normalize = TRUE,
  trim = 0.01,
  cut_off = 0.4,
  ...
)
}
\arguments{
\item{spec}{the spectral data, as an output from \code{get_2Dspectrum}}

\item{time}{the time to plot. Either an integer or the date.}

\item{normalize}{Should the spectrum be normalized to have maximum 1 before plotting}

\item{trim}{removes the values of the spectral density lower than this value}

\item{cut_off}{cut-off frequency above which the spectrum is not plotted}

\item{...}{currently unused}
}
\value{
a ggplot object
}
\description{
Plot a wave density 2D spectrum
}
\examples{
plot_2d_specta(resourcecodedata::rscd_2d_spectra, 1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{\%nin\%}
\alias{\%nin\%}
\title{Value Matching}
\usage{
x \%nin\% table
}
\arguments{
\item{x}{value to search}

\item{table}{table of values}
}
\value{
the opposite of x \%in\% table
}
\description{
Value Matching
}
\examples{
1:10 \%in\% c(1, 3, 5, 9)
1:10 \%nin\% c(1, 3, 5, 9)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{jonswap}
\alias{jonswap}
\title{JONWSAP spectrum}
\usage{
jonswap(
  hs = 5,
  tp = 15,
  fmax = resourcecodedata::rscd_freq,
  df = NULL,
  gam = 3.3
)
}
\arguments{
\item{hs}{Hs (default: 5m)}

\item{tp}{Period (default: 10s)}

\item{fmax}{higher frequency of the spectrum or
vector of frequencies (default to resourcecode frequency vector)}

\item{df}{frequency step (unused if fmax=vector of frequencies)}

\item{gam}{peak enhancement factor (default: 3.3)}
}
\value{
Density spectrum with corresponding parameters
}
\description{
Creates a JONWSAP density spectrum (one-sided), defined by its integral parameters.
}
\details{
Reference :
\itemize{
\item O.G.Houmb and T.Overvik, "Parametrization of Wave Spectra and Long Term
Joint Distribution of Wave Height and Period,"
in Proceedings, First International Conference
on Behaviour of Offshore Structures (BOSS), Trondheim 1976.
23rd International Towing Tank Conference, vol. II, pp. 544-551
\item ITTC Committee, 2002, "The Specialist Committee on Waves -
Final Report and Recommendations to the 23rd ITTC",
Proc. ITTC, vol. II, pp. 505-736.
}
}
\examples{
S1 <- jonswap(tp = 15)
S2 <- jonswap(tp = 15, fmax = 0.95, df = 0.003)
plot(S1, type = "l", ylim = c(0, 72))
lines(S2, col = "red")
abline(v = 1 / 15)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{fractional_day_of_year}
\alias{fractional_day_of_year}
\title{Compute Fractional Day of Year from POSIXct}
\usage{
fractional_day_of_year(datetime)
}
\arguments{
\item{datetime}{A POSIXct object or vector of POSIXct objects representing
date-time values. Must have a timezone attribute.}
}
\value{
A numeric vector of the same length as \code{datetime}, containing
fractional day of year values. January 1st at midnight corresponds to 0,
and each hour adds approximately 0.04167 (1/24) to the value.
}
\description{
Calculates the fractional day of year from a POSIXct datetime object.
The fractional day is zero-indexed, starting at 0 for January 1st at
midnight and ending at approximately 365.958 for December 31st at 23:00
in a non-leap year (or 366.958 in a leap year).
}
\details{
The function computes the time difference in hours between the input
datetime and midnight on January 1st of the same year, then divides by 24
to obtain fractional days. The calculation accounts for leap years
automatically.
}
\examples{
dates <- seq(
  from = as.POSIXct("2024-01-01 00:00:00", tz = "UTC"),
  to = as.POSIXct("2024-01-02 00:00:00", tz = "UTC"),
  by = "6 hours"
)
fractional_day_of_year(dates) # Returns 0.00, 0.25, 0.50, 0.75, 1.00

# End of year
dt_end <- as.POSIXct("2024-12-31 23:00:00", tz = "UTC")
fractional_day_of_year(dt_end) # Returns ~365.958

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{metconv2zmcomp}
\alias{metconv2zmcomp}
\title{Convert meteorological wind speed and direction to u/v components}
\usage{
metconv2zmcomp(speed, direction, names = c("uwnd", "vwnd"))
}
\arguments{
\item{speed}{Numeric vector of wind speeds.}

\item{direction}{Numeric vector of wind directions in degrees (0° = from north,
90° = from east, 180° = from south, 270° = from west).}

\item{names}{(optional) ames to construct the resulting data.frame.}
}
\value{
A data.frame with two columns:
\describe{
\item{u}{Zonal wind component (m/s), positive eastward.}
\item{v}{Meridional wind component (m/s), positive northward.}
}
}
\description{
Converts wind speed (magnitude) and direction (in degrees, meteorological
convention: direction from which the wind blows, measured clockwise from north)
into zonal (u) and meridional (v) components.
}
\examples{
# Example 1: North wind of 10 m/s (blowing southward)
metconv2zmcomp(10, 0)

# Example 2: East wind of 5 m/s (blowing westward)
metconv2zmcomp(5, 90)

# Example 3: South wind of 8 m/s (blowing northward)
metconv2zmcomp(8, 180)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spectral_data_download.R
\name{get_2d_spectrum}
\alias{get_2d_spectrum}
\title{Download the 2D spectrum data from IFREMER ftp}
\usage{
get_2d_spectrum(point, start = "1994-01-01", end = "1994-02-28")
}
\arguments{
\item{point}{the location name (string) requested.
Alternatively, the node number. The consistency is checked internally.}

\item{start}{the starting date (as a string). The consistency is checked internally.}

\item{end}{the ending date as a string}
}
\value{
A list with 9 elements:
\describe{
\item{longitude}{Longitude}
\item{latitude}{Latitude}
\item{frequency1}{Lower frequency}
\item{frequency2}{Upper frequency}
\item{ef}{Surface elevation variance spectral density}
\item{th1m}{Mean direction from first spectral moment}
\item{th2m}{Mean direction from second spectral moment}
\item{sth1m}{Mean directional spreading from first spectral moment}
\item{sth2m}{Mean directional spreading from second spectral moment}
\item{freq}{Central frequency}
\item{dir}{Directionnal bins}
\item{forcings}{A data.frame with 6 variables:
\describe{
\item{time}{Time}
\item{dpt}{Depth, positive downward}
\item{wnd}{Wind intensity, at 10m above sea level}
\item{wnddir}{Wind direction, comes from}
\item{cur}{Current intensity, at the surface}
\item{curdir}{Current direction, going to}
}}
\item{station}{Station name}
}
}
\description{
Download the 2D spectrum data from IFREMER ftp
}
\examples{
spec2D <- get_2d_spectrum("SEMREVO", start = "1994-01-01", end = "1994-02-28")
if(!is.null(spec2D)){
  image(spec2D$dir, spec2D$freq, spec2D$efth[, , 1],
    xlab = "Direction (°)",
    ylab = "Frequency (Hz"
  )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/specta_data_processing.R
\name{dispersion}
\alias{dispersion}
\title{Compute the dispersion relation of waves
Find \emph{k} s.t. (2.pi.f)^2 = g.k.tanh(k.d)}
\usage{
dispersion(frequencies, depth, iter_max = 200, tol = 1e-06)
}
\arguments{
\item{frequencies}{frequency vector}

\item{depth}{depth (m)}

\item{iter_max}{maximum number of iterations in the solver}

\item{tol}{tolerance for termination.}
}
\value{
the wave numbers (same size as frequencies)
}
\description{
Compute the dispersion relation of waves
Find \emph{k} s.t. (2.pi.f)^2 = g.k.tanh(k.d)
}
\examples{
freq <- seq(from = 0, to = 1, length.out = 100)
k1 <- dispersion(freq, depth = 1)
k10 <- dispersion(freq, depth = 10)
kInf <- dispersion(freq, depth = Inf)
plot(freq, k1, type = "l")
lines(freq, k10, col = "red")
lines(freq, kInf, col = "green")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{mean_direction}
\alias{mean_direction}
\title{Mean Direction}
\usage{
mean_direction(directions, weights = NULL)
}
\arguments{
\item{directions}{numeric vector of directions, in degree, 0° being the North}

\item{weights}{numeric vector, usually wind speed of wave height.}
}
\value{
The (weighted) mean of the values in \code{directions} is computed.
}
\description{
Function for computing the (weighted) arithmetic mean of
directional data in \strong{meteorological convention}.
}
\examples{
# Test with some wind directions (unweighted)
wind_directions <- c(10, 20, 350, 5, 15) # Directions mostly around North
mean_dir <- mean_direction(wind_directions)
cat("Mean wind direction (unweighted):", round(mean_dir, 1), "degrees\n")
wind_directions <- c(350, 10, 20, 340, 30) # Directions around North
wind_speeds <- c(15, 5, 2, 12, 3) # Higher speeds for directions closer to North
mean_dir_weighted <- mean_direction(wind_directions, wind_speeds)
cat("Mean wind direction (weighted):", round(mean_dir_weighted, 1), "degrees\n")

# Compare weighted vs unweighted for the same data
mean_dir_unweighted <- mean_direction(wind_directions)
cat("Same data unweighted:", round(mean_dir_unweighted, 1), "degrees\n")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{cut_directions}
\alias{cut_directions}
\title{Directional binning}
\usage{
cut_directions(directions, n_bins = 8, labels = NULL)
}
\arguments{
\item{directions}{vector of directions to be binned, in degree, 0° being the North.}

\item{n_bins}{number of bins, default: 8 sectors.}

\item{labels}{optional character vector giving the sectors names.}
}
\value{
a factor vector the same size as \code{directions} with the values binned into sectors.
}
\description{
Cuts direction vector into directional bins
}
\examples{
# Example usage and demonstration
set.seed(123)
directions <- runif(20, 0, 360)

# Test with different numbers of bins
cat("Original directions:\n")
print(round(directions, 1))

cat("\n8 bins (default):\n")
bins_8 <- cut_directions(directions, n_bins = 8)
print(bins_8)

cat("\n4 bins:\n")
bins_4 <- cut_directions(directions, n_bins = 4)
print(bins_4)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/weather_windows.R
\name{weather_windows}
\alias{weather_windows}
\title{Compute Weather Windows}
\usage{
weather_windows(
  valid_periods,
  window_length,
  allow_overlap = TRUE,
  time_step = 3600
)
}
\arguments{
\item{valid_periods}{A data frame with a 'time' column (POSIXct).}

\item{window_length}{Minimum window duration (hours).}

\item{allow_overlap}{Logical; If TRUE, the algorithm searches for window,
if a window is found, search of next window will start from the end of the previous window.
If FALSE, it uses continuous window search: The algorithm searches for window starting from
every time step that meets the criteria.}

\item{time_step}{Expected time step between consecutive timestamps (seconds).}
}
\value{
POSIXct vector of detected window start times.
}
\description{
Computes and returns start date of each weather window, implemented in C++ for speed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{cut_seasons}
\alias{cut_seasons}
\title{Get season from date time object}
\usage{
cut_seasons(
  datetime,
  definition = "meteorological",
  hemisphere = "northern",
  labels = NULL
)
}
\arguments{
\item{datetime}{a POSIXct vector from with the season is constructed}

\item{definition}{the definition used to compute the season. See details section.}

\item{hemisphere}{in the Southern hemisphere, seasons are reversed compared to the Northern one.}

\item{labels}{optional, a character vector of length fours with the seasons' names.}
}
\value{
a Factor vector with 4 levels depending on the definitions (and labels if provided)
}
\description{
Get season from date time object
}
\details{
Available Definitions:
\itemize{
\item meteorological: Standard seasons (Dec-Feb = Winter, etc.)
\item astronomical: Based on equinoxes/solstices
\item djf: Dec-Jan-Feb, Mar-Apr-May, Jun-Jul-Aug, Sep-Oct-Nov
\item jfm: Jan-Feb-Mar, Apr-May-Jun, Jul-Aug-Sep, Oct-Nov-Dec
\item fma: Feb-Mar-Apr, May-Jun-Jul, Aug-Sep-Oct, Nov-Dec-Jan
\item amj, jas, ond: Alternative starting points for quarterly seasons
}
}
\examples{
dates <- seq(
  from = as.POSIXct("2023-01-15"),
  to = as.POSIXct("2023-12-15"),
  by = "month"
)
cut_seasons(dates)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{zmcomp2metconv}
\alias{zmcomp2metconv}
\title{Convert u/v to meteorological wind speed and direction}
\usage{
zmcomp2metconv(u, v = NULL, names = c("wspd", "wdir"))
}
\arguments{
\item{u}{zonal velocity (1D vector) or matrix with zonal and meridional velocity (Nx2 matrix)}

\item{v}{meridional velocity (1D vector)}

\item{names}{names to construct the resulting data.frame}
}
\value{
a Nx2 data.frame with the norm and direction (meteorological convention)
}
\description{
Converts wind or current zonal and meridional velocity components to
magnitude and direction according to meteorological convention.
}
\examples{
u <- matrix(rnorm(200), nrow = 100, ncol = 2)
vdir <- zmcomp2metconv(u)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/specta_data_processing.R
\name{convert_spectrum_2d1d}
\alias{convert_spectrum_2d1d}
\title{Converts a 2D spectrum time series to a 1D spectrum}
\usage{
convert_spectrum_2d1d(spec, ...)
}
\arguments{
\item{spec}{a structure with the needed fields, as an output from 'get_2Dspectrum' for example}

\item{...}{unused yet}
}
\value{
a structure comparable to 'get_1Dspectrum'.
}
\description{
Converts a 2D spectrum time series to a 1D spectrum
}
\examples{
spec <- resourcecodedata::rscd_2d_spectra
spec1D_RSCD <- resourcecodedata::rscd_1d_spectra
spec1D <- convert_spectrum_2d1d(spec)
# Check the differences, should be low
max(abs(spec1D_RSCD$ef - spec1D$ef))

# Plot the different spectrum
plot(spec1D$freq, spec1D$ef[, 1], type = "l", log = "xy")
lines(spec1D_RSCD$freq, spec1D_RSCD$ef[, 1], col = "red")

# Images
lims <- c(0, 360)
r <- as.POSIXct(round(range(spec1D$forcings$time), "hours"))
oldpar <- par(mfcol = c(2, 1))
image(spec1D$forcings$time, spec1D$freq, t(spec1D$th1m),
  zlim = lims,
  xlab = "Time",
  ylab = "Freq (Hz)",
  xaxt = "n",
  main = "Directionnal spreading"
)
axis.POSIXct(1, spec1D$forcings$time,
  at = seq(r[1], r[2], by = "week"),
  format = "\%Y-\%m-\%d",
  las = 2
)
image(spec1D_RSCD$forcings$time, spec1D_RSCD$freq, t(spec1D_RSCD$th1m),
  zlim = lims,
  xlab = "Time",
  ylab = "Freq (Hz)",
  xaxt = "n"
)
axis.POSIXct(1, spec1D$forcings$time,
  at = seq(r[1], r[2], by = "week"),
  format = "\%Y-\%m-\%d",
  las = 2
)
par(oldpar)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plots.R
\name{rscd_mapplot}
\alias{rscd_mapplot}
\title{Create a map of the provided variable on the RESOURCECODE field grid}
\usage{
rscd_mapplot(
  z,
  name = "Depth (m)",
  zlim = NULL,
  palette = "YlOrRd",
  direction = 1,
  transform = "identity"
)
}
\arguments{
\item{z}{the data ro plot: a vector of the same size as the grid (328,030 rows).}

\item{name}{name of the variable plored, to be included in the legend.}

\item{zlim}{limits of the scale. See \code{\link[ggplot2]{continuous_scale}} for details.}

\item{palette}{If a string, will use that named palette.
See \code{\link[ggplot2]{scale_colour_brewer}} for other options.}

\item{direction}{Sets the order of colours in the scale.
See \code{\link[ggplot2]{scale_colour_brewer}} for details.}

\item{transform}{Transformation to apply to the color scale.
See \code{\link[ggplot2]{continuous_scale}} for details.}
}
\value{
a ggplot2 object
}
\description{
Create a map of the provided variable on the RESOURCECODE field grid
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
rscd_mapplot(resourcecodedata::rscd_field$depth)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/orbital_speed.R
\name{compute_orbital_speeds}
\alias{compute_orbital_speeds}
\title{Compute the orbital speed at a given depth from the wave elevation 1D spectra}
\usage{
compute_orbital_speeds(spec, freq, z = 0, depth = Inf, output_speeds = FALSE)
}
\arguments{
\item{spec}{1D spectral data: TxM matrix}

\item{freq}{the M frequencies}

\item{z}{distance above the floor at which we want the orbital speed (single numeric)}

\item{depth}{depth time series (vector length T. Recycled if a single value is given)}

\item{output_speeds}{TRUE if the spectral speed are needed. Otherwise, returns the RMS (default)}
}
\value{
depending on spec, a list with the spectral velocities for each component
(if output_speeds==FALSE) or a data.frame with a time series of horizontal and vertical
components of (spectral) orbital speed.
}
\description{
Compute the orbital speed at a given depth from the wave elevation 1D spectra
}
\examples{
# Compute orbital speed for varying Hs
S <- t(sapply(1:10, function(h) {
  jonswap(h)$spec
}))
orb_speeds <- compute_orbital_speeds(S, resourcecodedata::rscd_freq, depth = 100, z = 10)
plot(1:10, orb_speeds[, 1],
  type = "l",
  ylim = range(orb_speeds),
  xlab = "Hs (m)",
  ylab = "Orbital speed RMS (m/s)"
)
lines(1:10, orb_speeds[, 2], type = "l", col = "red")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/specta_data_processing.R
\name{compute_sea_state_1d_spectrum}
\alias{compute_sea_state_1d_spectrum}
\title{Compute sea_state parameter from wave spectrum}
\usage{
compute_sea_state_1d_spectrum(spec, ...)
}
\arguments{
\item{spec}{1D spectrum data, e.g. from \code{get_1Dspectrum}}

\item{...}{currently unused}
}
\value{
a tibble with the sea-state parameters computed from the time series of 1D spectrum
}
\description{
Compute sea_state parameter from wave spectrum
}
\examples{
\dontshow{if (identical(tolower(Sys.getenv("NOT_CRAN")), "true")) withAutoprint(\{ # examplesIf}
rscd_params <- get_parameters(
  node = "134865",
  start = "1994-01-01",
  end = "1994-01-31 23:00:00",
  parameters = c("hs", "tp")
)
spec <- resourcecodedata::rscd_1d_spectra
param_calc <- compute_sea_state_1d_spectrum(spec)
oldpar <- par(mfcol = c(2, 1))
plot(param_calc$time, param_calc$hs, type = "l", xlab = "Time", ylab = "Hs (m)")
lines(rscd_params$time, rscd_params$hs, col = "red")
plot(param_calc$time, param_calc$tp, type = "l", xlab = "Time", ylab = "Tp (s)")
lines(rscd_params$time, rscd_params$tp, col = "red")
par(oldpar)
\dontshow{\}) # examplesIf}
}
